# select_parser.py
# Copyright 2010,2019 Paul McGuire
#
# a simple SELECT statement parser, taken from SQLite's SELECT statement
# definition at https://www.sqlite.org/lang_select.html
#
from pyparsing import *
ParserElement.enablePackrat()

LPAR,RPAR,COMMA = map(Suppress,"(),")
DOT,STAR = map(Literal, ".*")
select_stmt = Forward().setName("select statement")

# keywords
(UNION, ALL, AND, INTERSECT, EXCEPT, COLLATE, ASC, DESC, ON, USING, NATURAL, INNER,
 CROSS, LEFT, OUTER, JOIN, AS, INDEXED, NOT, SELECT, DISTINCT, FROM, WHERE, GROUP, BY,
 HAVING, ORDER, BY, LIMIT, OFFSET, OR) =  map(CaselessKeyword, """UNION, ALL, AND, INTERSECT,
 EXCEPT, COLLATE, ASC, DESC, ON, USING, NATURAL, INNER, CROSS, LEFT, OUTER, JOIN, AS, INDEXED, NOT, SELECT,
 DISTINCT, FROM, WHERE, GROUP, BY, HAVING, ORDER, BY, LIMIT, OFFSET, OR""".replace(",","").split())
(CAST, ISNULL, NOTNULL, NULL, IS, BETWEEN, ELSE, END, CASE, WHEN, THEN, EXISTS, IN, LIKE, GLOB, REGEXP,
 MATCH, ESCAPE, CURRENT_TIME, CURRENT_DATE, CURRENT_TIMESTAMP) = map(CaselessKeyword,
 """CAST, ISNULL, NOTNULL, NULL, IS, BETWEEN, ELSE, END, CASE, WHEN, THEN, EXISTS, IN, LIKE, GLOB, 
 REGEXP, MATCH, ESCAPE, CURRENT_TIME, CURRENT_DATE, CURRENT_TIMESTAMP""".replace(",","").split())
keyword = MatchFirst((UNION, ALL, INTERSECT, EXCEPT, COLLATE, ASC, DESC, ON, USING, NATURAL, INNER,
                      CROSS, LEFT, OUTER, JOIN, AS, INDEXED, NOT, SELECT, DISTINCT, FROM, WHERE, GROUP, BY,
                      HAVING, ORDER, BY, LIMIT, OFFSET, CAST, ISNULL, NOTNULL, NULL, IS, BETWEEN, ELSE, END,
                      CASE, WHEN, THEN, EXISTS, COLLATE, IN, LIKE, GLOB, REGEXP, MATCH, ESCAPE,
                      CURRENT_TIME, CURRENT_DATE, CURRENT_TIMESTAMP))

identifier = ~keyword + Word(alphas, alphanums+"_")
collation_name = identifier.copy()
column_name = identifier.copy()
column_alias = identifier.copy()
table_name = identifier.copy()
table_alias = identifier.copy()
index_name = identifier.copy()
function_name = identifier.copy()
parameter_name = identifier.copy()
database_name = identifier.copy()

# expression
expr = Forward().setName("expression")

integer = Regex(r"[+-]?\d+")
numeric_literal = Regex(r"\d+(\.\d*)?([eE][+-]?\d+)?")
string_literal = QuotedString("'")
blob_literal = Regex(r"[xX]'[0-9A-Fa-f]+'")
literal_value = (
        numeric_literal
        | string_literal
        | blob_literal
        | NULL
        | CURRENT_TIME
        | CURRENT_DATE
        | CURRENT_TIMESTAMP
        )
bind_parameter = (
    Word("?",nums)
    | Combine(oneOf(": @ $") + parameter_name)
    )
type_name = oneOf("TEXT REAL INTEGER BLOB NULL")

expr_term = (
    CAST + LPAR + expr + AS + type_name + RPAR
    | EXISTS + LPAR + select_stmt + RPAR
    | function_name.setName("function_name") + LPAR + Optional(STAR | delimitedList(expr)) + RPAR
    | literal_value
    | bind_parameter
    | Group(identifier('col_db') + DOT + identifier('col_tab') + DOT + identifier('col'))
    | Group(identifier('col_tab') + DOT + identifier('col'))
    | Group(identifier('col'))
    )

UNARY,BINARY,TERNARY=1,2,3
expr << infixNotation(expr_term,
    [
    (oneOf('- + ~') | NOT, UNARY, opAssoc.RIGHT),
    (ISNULL | NOTNULL | NOT + NULL, UNARY, opAssoc.LEFT),
    ('||', BINARY, opAssoc.LEFT),
    (oneOf('* / %'), BINARY, opAssoc.LEFT),
    (oneOf('+ -'), BINARY, opAssoc.LEFT),
    (oneOf('<< >> & |'), BINARY, opAssoc.LEFT),
    (oneOf('< <= > >='), BINARY, opAssoc.LEFT),
    (oneOf('= == != <>') | IS | IN | LIKE | GLOB | MATCH | REGEXP, BINARY, opAssoc.LEFT),
    ((BETWEEN,AND), TERNARY, opAssoc.LEFT),
    (IN + LPAR + Group(select_stmt | delimitedList(expr)) + RPAR, UNARY, opAssoc.LEFT),
    (AND, BINARY, opAssoc.LEFT),
    (OR, BINARY, opAssoc.LEFT),
    ])

compound_operator = (UNION + Optional(ALL) | INTERSECT | EXCEPT)

ordering_term = Group(expr('order_key')
                      + Optional(COLLATE + collation_name('collate'))
                      + Optional(ASC | DESC)('direction'))

join_constraint = Group(Optional(ON + expr | USING + LPAR + Group(delimitedList(column_name)) + RPAR))

join_op = COMMA | Group(Optional(NATURAL) + Optional(INNER | CROSS | LEFT + OUTER | LEFT | OUTER) + JOIN)

join_source = Forward()
single_source = (
                Group(database_name("database") + DOT + table_name("table*")
                      | table_name("table*"))
                    + Optional(Optional(AS) + table_alias("table_alias*"))
                    + Optional(INDEXED + BY + index_name("name") | NOT + INDEXED)("index")
                | (LPAR + select_stmt + RPAR + Optional(Optional(AS) + table_alias))
                | (LPAR + join_source + RPAR)
                )

join_source <<= (Group(single_source + OneOrMore(join_op + single_source + join_constraint))
                 | single_source)

# result_column = "*" | table_name + "." + "*" | Group(expr + Optional(Optional(AS) + column_alias))
result_column = Group(STAR('col')
                      | table_name('col_table') + DOT + STAR('col')
                      | expr('col') + Optional(Optional(AS) + column_alias('alias'))
                      )

select_core = (SELECT + Optional(DISTINCT | ALL) + Group(delimitedList(result_column))("columns")
               + Optional(FROM + join_source("from*"))
               + Optional(WHERE + expr("where_expr"))
               + Optional(GROUP + BY + Group(delimitedList(ordering_term))("group_by_terms")
                          + Optional(HAVING + expr("having_expr"))))

select_stmt << (select_core + ZeroOrMore(compound_operator + select_core)
                + Optional(ORDER + BY + Group(delimitedList(ordering_term))("order_by_terms"))
                + Optional(LIMIT + (Group(expr + OFFSET + expr) | Group(expr + COMMA + expr) | expr)("limit"))
                )

tests = """\
    select * from xyzzy where z > 100
    select * from xyzzy where z > 100 order by zz
    select * from xyzzy
    select z.* from xyzzy
    select a, b from test_table where 1=1 and b='yes'
    select a, b from test_table where 1=1 and b in (select bb from foo)
    select z.a, b from test_table where 1=1 and b in (select bb from foo)
    select z.a, b from test_table where 1=1 and b in (select bb from foo) order by b,c desc,d
    select z.a, b from test_table left join test2_table where 1=1 and b in (select bb from foo)
    select a, db.table.b as BBB from db.table where 1=1 and BBB='yes'
    select a, db.table.b as BBB from test_table,db.table where 1=1 and BBB='yes'
    select a, db.table.b as BBB from test_table,db.table where 1=1 and BBB='yes' limit 50
    select a, b from test_table where (1=1 or 2=3) and b='yes' group by zx having b=2 order by 1
    SELECT emp.ename as e FROM scott.employee as emp
    SELECT ename as e, fname as f FROM scott.employee as emp
    SELECT emp.eid, fname,lname FROM scott.employee as emp
    SELECT ename, lname, emp.eid FROM scott.employee as emp
    select emp.salary * (1.0 + emp.bonus) as salary_plus_bonus from scott.employee as emp
"""
select_stmt.runTests(tests)



from Aspidites._vendor.pyrsistent import pset, s
import pytest
import pickle

def test_key_is_tuple():
    with pytest.raises(KeyError):
        pset().remove((1,1))

def test_literalish_works():
    assert s() is pset()
    assert s(1, 2) == pset([1, 2])


def test_supports_hash():
    assert hash(s(1, 2)) == hash(s(1, 2))


def test_empty_truthiness():
    assert s(1)
    assert not s()


def test_contains_elements_that_it_was_initialized_with():
    initial = [1, 2, 3]
    s = pset(initial)

    assert set(s) == set(initial)
    assert len(s) == len(set(initial))


def test_is_immutable():
    s1 = pset([1])
    s2 = s1.add(2)

    assert s1 == pset([1])
    assert s2 == pset([1, 2])

    s3 = s2.remove(1)
    assert s2 == pset([1, 2])
    assert s3 == pset([2])


def test_remove_when_not_present():
    s1 = s(1, 2, 3)
    with pytest.raises(KeyError):
        s1.remove(4)


def test_discard():
    s1 = s(1, 2, 3)
    assert s1.discard(3) == s(1, 2)
    assert s1.discard(4) is s1


def test_is_iterable():
    assert sum(pset([1, 2, 3])) == 6


def test_contains():
    s = pset([1, 2, 3])

    assert 2 in s
    assert 4 not in s


def test_supports_set_operations():
    s1 = pset([1, 2, 3])
    s2 = pset([3, 4, 5])

    assert s1 | s2 == s(1, 2, 3, 4, 5)
    assert s1.union(s2) == s1 | s2

    assert s1 & s2 == s(3)
    assert s1.intersection(s2) == s1 & s2

    assert s1 - s2 == s(1, 2)
    assert s1.difference(s2) == s1 - s2

    assert s1 ^ s2 == s(1, 2, 4, 5)
    assert s1.symmetric_difference(s2) == s1 ^ s2


def test_supports_set_comparisons():
    s1 = s(1, 2, 3)
    s3 = s(1, 2)
    s4 = s(1, 2, 3)

    assert s(1, 2, 3, 3, 5) == s(1, 2, 3, 5)
    assert s1 != s3

    assert s3 < s1
    assert s3 <= s1
    assert s3 <= s4

    assert s1 > s3
    assert s1 >= s3
    assert s4 >= s3


def test_str():
    rep = str(pset([1, 2, 3]))
    assert rep == "pset([1, 2, 3])"


def test_is_disjoint():
    s1 = pset([1, 2, 3])
    s2 = pset([3, 4, 5])
    s3 = pset([4, 5])

    assert not s1.isdisjoint(s2)
    assert s1.isdisjoint(s3)


def test_evolver_simple_add():
    x = s(1, 2, 3)
    e = x.evolver()
    assert not e.is_dirty()

    e.add(4)
    assert e.is_dirty()

    x2 = e.persistent()
    assert not e.is_dirty()
    assert x2 == s(1, 2, 3, 4)
    assert x == s(1, 2, 3)

def test_evolver_simple_remove():
    x = s(1, 2, 3)
    e = x.evolver()
    e.remove(2)

    x2 = e.persistent()
    assert x2 == s(1, 3)
    assert x == s(1, 2, 3)


def test_evolver_no_update_produces_same_pset():
    x = s(1, 2, 3)
    e = x.evolver()
    assert e.persistent() is x


def test_evolver_len():
    x = s(1, 2, 3)
    e = x.evolver()
    assert len(e) == 3


def test_copy_returns_reference_to_self():
    s1 = s(10)
    assert s1.copy() is s1


def test_pickling_empty_set():
    assert pickle.loads(pickle.dumps(s(), -1)) == s()


def test_pickling_non_empty_map():
    assert pickle.loads(pickle.dumps(s(1, 2), -1)) == s(1, 2)


def test_supports_weakref():
    import weakref
    weakref.ref(s(1))


def test_update():
    assert s(1, 2, 3).update([3, 4, 4, 5]) == s(1, 2, 3, 4, 5)


def test_update_no_elements():
    s1 = s(1, 2)
    assert s1.update([]) is s1


def test_iterable():
    """
    PSets can be created from iterables even though they can't be len() hinted.
    """

    assert pset(iter("a")) == pset(iter("a"))



import pickle
import pytest
from Aspidites._vendor.pyrsistent import pdeque, dq


def test_basic_right_and_left():
    x = pdeque([1, 2])

    assert x.right == 2
    assert x.left == 1
    assert len(x) == 2


def test_construction_with_maxlen():
    assert pdeque([1, 2, 3, 4], maxlen=2) == pdeque([3, 4])
    assert pdeque([1, 2, 3, 4], maxlen=4) == pdeque([1, 2, 3, 4])
    assert pdeque([], maxlen=2) == pdeque()


def test_construction_with_invalid_maxlen():
    with pytest.raises(TypeError):
        pdeque([], maxlen='foo')

    with pytest.raises(ValueError):
        pdeque([], maxlen=-3)


def test_pop():
    x = pdeque([1, 2, 3, 4]).pop()
    assert x.right == 3
    assert x.left == 1

    x = x.pop()
    assert x.right == 2
    assert x.left == 1

    x = x.pop()
    assert x.right == 1
    assert x.left == 1

    x = x.pop()
    assert x == pdeque()

    x = pdeque([1, 2]).pop()
    assert x == pdeque([1])

    x = x.pop()
    assert x == pdeque()

    assert pdeque().append(1).pop() == pdeque()
    assert pdeque().appendleft(1).pop() == pdeque()


def test_pop_multiple():
    assert pdeque([1, 2, 3, 4]).pop(3) == pdeque([1])
    assert pdeque([1, 2]).pop(3) == pdeque()


def test_pop_with_negative_index():
    assert pdeque([1, 2, 3]).pop(-1) == pdeque([1, 2, 3]).popleft(1)
    assert pdeque([1, 2, 3]).popleft(-1) == pdeque([1, 2, 3]).pop(1)


def test_popleft():
    x = pdeque([1, 2, 3, 4]).popleft()
    assert x.left == 2
    assert x.right == 4

    x = x.popleft()
    assert x.left == 3
    assert x.right == 4

    x = x.popleft()
    assert x.right == 4
    assert x.left == 4

    x = x.popleft()
    assert x == pdeque()

    x = pdeque([1, 2]).popleft()
    assert x == pdeque([2])

    x = x.popleft()
    assert x == pdeque()

    assert pdeque().append(1).popleft() == pdeque()
    assert pdeque().appendleft(1).popleft() == pdeque()


def test_popleft_multiple():
    assert pdeque([1, 2, 3, 4]).popleft(3) == pdeque([4])


def test_left_on_empty_deque():
    with pytest.raises(IndexError):
        pdeque().left


def test_right_on_empty_deque():
    with pytest.raises(IndexError):
        pdeque().right


def test_pop_empty_deque_returns_empty_deque():
    # The other option is to throw an index error, this is what feels best for now though
    assert pdeque().pop() == pdeque()
    assert pdeque().popleft() == pdeque()


def test_str():
    assert str(pdeque([1, 2, 3])) == 'pdeque([1, 2, 3])'
    assert str(pdeque([])) == 'pdeque([])'
    assert str(pdeque([1, 2], maxlen=4)) == 'pdeque([1, 2], maxlen=4)'


def test_append():
    assert pdeque([1, 2]).append(3).append(4) == pdeque([1, 2, 3, 4])


def test_append_with_maxlen():
    assert pdeque([1, 2], maxlen=2).append(3).append(4) == pdeque([3, 4])
    assert pdeque([1, 2], maxlen=3).append(3).append(4) == pdeque([2, 3, 4])
    assert pdeque([], maxlen=0).append(1) == pdeque()


def test_appendleft():
    assert pdeque([2, 1]).appendleft(3).appendleft(4) == pdeque([4, 3, 2, 1])


def test_appendleft_with_maxlen():
    assert pdeque([2, 1], maxlen=2).appendleft(3).appendleft(4) == pdeque([4, 3])
    assert pdeque([2, 1], maxlen=3).appendleft(3).appendleft(4) == pdeque([4, 3, 2])
    assert pdeque([], maxlen=0).appendleft(1) == pdeque()


def test_extend():
    assert pdeque([1, 2]).extend([3, 4]) == pdeque([1, 2, 3, 4])


def test_extend_with_maxlen():
    assert pdeque([1, 2], maxlen=3).extend([3, 4]) == pdeque([2, 3, 4])
    assert pdeque([1, 2], maxlen=2).extend([3, 4]) == pdeque([3, 4])
    assert pdeque([], maxlen=2).extend([1, 2]) == pdeque([1, 2])
    assert pdeque([], maxlen=0).extend([1, 2]) == pdeque([])


def test_extendleft():
    assert pdeque([2, 1]).extendleft([3, 4]) == pdeque([4, 3, 2, 1])


def test_extendleft_with_maxlen():
    assert pdeque([1, 2], maxlen=3).extendleft([3, 4]) == pdeque([4, 3, 1])
    assert pdeque([1, 2], maxlen=2).extendleft([3, 4]) == pdeque([4, 3])
    assert pdeque([], maxlen=2).extendleft([1, 2]) == pdeque([2, 1])
    assert pdeque([], maxlen=0).extendleft([1, 2]) == pdeque([])


def test_count():
    x = pdeque([1, 2, 3, 2, 1])
    assert x.count(1) == 2
    assert x.count(2) == 2


def test_remove():
    assert pdeque([1, 2, 3, 4]).remove(2) == pdeque([1, 3, 4])
    assert pdeque([1, 2, 3, 4]).remove(4) == pdeque([1, 2, 3])

    # Right list must be reversed before removing element
    assert pdeque([1, 2, 3, 3, 4, 5, 4, 6]).remove(4) == pdeque([1, 2, 3, 3, 5, 4, 6])


def test_remove_element_missing():
    with pytest.raises(ValueError):
        pdeque().remove(2)

    with pytest.raises(ValueError):
        pdeque([1, 2, 3]).remove(4)


def test_reverse():
    assert pdeque([1, 2, 3, 4]).reverse() == pdeque([4, 3, 2, 1])
    assert pdeque().reverse() == pdeque()


def test_rotate_right():
    assert pdeque([1, 2, 3, 4, 5]).rotate(2) == pdeque([4, 5, 1, 2, 3])
    assert pdeque([1, 2]).rotate(0) == pdeque([1, 2])
    assert pdeque().rotate(2) == pdeque()


def test_rotate_left():
    assert pdeque([1, 2, 3, 4, 5]).rotate(-2) == pdeque([3, 4, 5, 1, 2])
    assert pdeque().rotate(-2) == pdeque()


def test_set_maxlen():
    x = pdeque([], maxlen=4)
    assert x.maxlen == 4

    with pytest.raises(AttributeError):
        x.maxlen = 5


def test_comparison():
    small = pdeque([1, 2])
    large = pdeque([1, 2, 3])

    assert small < large
    assert large > small
    assert not small > large
    assert not large < small
    assert large != small

    # Not equal to other types
    assert small != [1, 2]


def test_pickling():
    input = pdeque([1, 2, 3], maxlen=5)
    output = pickle.loads(pickle.dumps(input, -1))

    assert output == input
    assert output.maxlen == input.maxlen


def test_indexing():
    assert pdeque([1, 2, 3])[0] == 1
    assert pdeque([1, 2, 3])[1] == 2
    assert pdeque([1, 2, 3])[2] == 3
    assert pdeque([1, 2, 3])[-1] == 3
    assert pdeque([1, 2, 3])[-2] == 2
    assert pdeque([1, 2, 3])[-3] == 1


def test_one_element_indexing():
    assert pdeque([2])[0] == 2
    assert pdeque([2])[-1] == 2


def test_empty_indexing():
    with pytest.raises(IndexError):
        assert pdeque([])[0] == 1


def test_indexing_out_of_range():
    with pytest.raises(IndexError):
        pdeque([1, 2, 3])[-4]

    with pytest.raises(IndexError):
        pdeque([1, 2, 3])[3]

    with pytest.raises(IndexError):
        pdeque([2])[-2]


def test_indexing_invalid_type():
    with pytest.raises(TypeError) as e:
        pdeque([1, 2, 3])['foo']

    assert 'cannot be interpreted' in str(e.value)


def test_slicing():
    assert pdeque([1, 2, 3])[1:2] == pdeque([2])
    assert pdeque([1, 2, 3])[2:1] == pdeque([])
    assert pdeque([1, 2, 3])[-2:-1] == pdeque([2])
    assert pdeque([1, 2, 3])[::2] == pdeque([1, 3])


def test_hashing():
    assert hash(pdeque([1, 2, 3])) == hash(pdeque().append(1).append(2).append(3))


def test_index():
    assert pdeque([1, 2, 3]).index(3) == 2


def test_literalish():
    assert dq(1, 2, 3) == pdeque([1, 2, 3])


def test_supports_weakref():
    import weakref
    weakref.ref(dq(1, 2))


def test_iterable():
    """
    PDeques can be created from iterables even though they can't be len()
    hinted.
    """

    assert pdeque(iter("a")) == pdeque(iter("a"))



import os
import re
import sys
import warnings
from pathlib import Path
from keyword import iskeyword
from string import ascii_letters, digits, ascii_lowercase
import hypothesis
import pytest as pt
from hypothesis import given, assume, strategies as st
from Aspidites._vendor.contracts import contract as __contract

try:
    from numpy import inf, nan, isinf, isnan
except ImportError:
    from math import inf, nan, isinf, isnan

from .._vendor.contracts import ContractNotRespected
from ..__main__ import get_cy_kwargs, main
from ..parser.parser import parse_module, func_def, arith_expr, list_item, collection_trigrams
from ..templates import woma_template
from ..monads import Maybe as __maybe, Surely
from ..math import Undefined as __undefined, SafeDiv as __safeDiv, SafeExp as __safeExp, SafeMod as __safeMod, SafeFloorDiv as __safeFloorDiv, SafeUnaryAdd as __safeUnaryAdd, SafeUnarySub as __safeUnarySub, SafeFactorial as __safeFactorial
from ..compiler import Compiler, CompilerArgs

docker = os.getenv("ASPIDITES_DOCKER_BUILD")

woma_file = str(Path('examples/examples.wom'))

valid_id = [c for c in ascii_letters + digits + '_']


@pt.fixture(autouse=True)
def inject_config(request):
    return request.config.rootpath


def setup_code(inject_config):
    if Path(woma_file).exists():
        code_ = parse_module(open(woma_file, 'r').read())
    else:
        try:
            code_ = parse_module(open(str(Path("Aspidites/tests") / woma_file), 'r').read())
            warnings.warn("Aspidites is being tested in source mode")
            os.environ['ASPIDITES_SOURCE_MODE'] = "TRUE"
        except FileNotFoundError:
            code_ = parse_module(open(Path(inject_config) / woma_file, 'r').read())
    return code_


def test_compile_module(inject_config):
    try:
        compile(woma_template.substitute(code='\n'.join(setup_code(inject_config))), '', 'exec')
    except FileNotFoundError:  # ????
        compile(woma_template.substitute(code='\n'.join(os.path.join(inject_config(), woma_file))), '',
                'exec')


def test_cli_help_exit():
    with pt.raises(SystemExit):
        main(['Aspidites', '-h'])


def test_cli_no_arg_exit():
    with pt.raises(SystemExit):
        main(['Aspidites', ''])


def test_cli_no_target_exit():
    with pt.raises(SystemExit):
        main(['Aspidites', '-fpc'])


def test_cli_examples():
    python_file_ = Path('examples/compiled.py')
    python_file = python_file_ if Path(woma_file).exists() else (Path('Aspidites/tests') / python_file_)
    with pt.raises(SystemExit):
        main(['Aspidites', woma_file, f'-o={python_file}', '-c'])


@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much], deadline=None)
@hypothesis.given(st.text([c for c in ascii_letters]),
                  st.text([c for c in ascii_lowercase]),
                  st.integers(),
                  st.integers(1, 10))
def test_parse_func_def(w, x, y, z):
    assume(str(w).isidentifier() and str(x).isidentifier())
    assume(not iskeyword(w) and not iskeyword(x))
    assume(w is not x)
    assume(y > 0)
    f = f"({w}({x} = {y} -> int)) int\n    <*>{x}"
    exec(''.join(func_def.parseString(f)))
    args = ''
    for i in range(z):
        args += f"{x + str(i)} = {y} -> int;"
    args = args[:-1]
    f2 = f"({w}({x} = {y} -> int; {args})) int\n    <*>{x}"
    print(f2)
    exec(''.join(func_def.parseString(f2)))


@hypothesis.settings(deadline=None)
@hypothesis.given(st.text([c for c in ascii_letters]),
                  st.integers(),
                  st.integers(min_value=0))
def test_parse_literals(v, x, y):
    assert ''.join(list_item.parseString(f"'{v}'")) == f"'{v}'"
    assert ''.join(list_item.parseString(f'"{v}"')) == f'"{v}"'
    assert ''.join(list_item.parseString('/0')) == '__undefined()'
    assert ''.join(list_item.parseString('True')) == 'True'
    assert ''.join(list_item.parseString('False')) == 'False'
    assert ''.join(list_item.parseString(f'{x}.{y}')) == f'{x}.{y}'
    assert ''.join(list_item.parseString(f'{y}+{x}j')) == f'{y}+{x}j'


@hypothesis.settings(deadline=None)
@hypothesis.given(st.text([c for c in ascii_letters]),
                  st.integers(),
                  st.integers(min_value=0))
def test_set_remove(v, x, y):
    assert ''.join(collection_trigrams.parseString(f"{{{y},{x}}}[*]{y}")) == f"__pset({{{y}, {x}}}).remove({y})"


# @hypothesis.settings(deadline=None)
# @hypothesis.given(st.text([c for c in ascii_letters], min_size=1),
#                   st.integers(),
#                   st.integers(min_value=0))
# def test_list_set(v, x, y):
#     assert ''.join(collection_trigrams.parseString(f"{v}[$]{x},{y}")) == f"{v}.set({x},{y})"
#     assert ''.join(collection_trigrams.parseString(f"{v}[$]{x},{v}")) == f"{v}.set({x},{v})"


@hypothesis.settings(deadline=None)
@hypothesis.given(st.sets(st.integers()),
                  st.lists(st.integers()),
                  st.text([c for c in ascii_letters]),
                  st.text([c for c in ascii_lowercase]),
                  st.integers(),
                  st.integers(min_value=0))
def test_parse_collections(t, u, v, w, x, y):
    assert ''.join(list_item.parseString(f"{{'a': ({y}+5), '{w}': 8, '{v}': True, {x}: None, 'd': 6**2*5+3}}")) \
           == f"__pmap({{'a': ({y}+5), '{w}': 8, '{v}': True, {x}: None, 'd': __maybe(__safeExp, 6, 2*5+3, )()}})"
    assert ''.join(list_item.parseString(f"{{'{v}', '{w}'}}")) == f"__pset({{'{v}', '{w}'}})"
    assert ''.join(list_item.parseString(f"[{u}, 4, 6, {x}, {y}]")) == f"__pvector([__pvector({u}), 4, 6, {x}, {y}])"


@hypothesis.settings(deadline=None)
@hypothesis.given(st.text([c for c in ascii_letters]),
                  st.text([c for c in ascii_lowercase]),
                  st.integers(),
                  st.integers(min_value=0))
def test_parse_evolvers(w, x, y, z):
    assert ''.join(list_item.parseString("{'a': (3+5), 'b': 8, 'c': True, 4: None, 'd': 6**2*5+3}...")) \
           == "__pmap({'a': (3+5), 'b': 8, 'c': True, 4: None, 'd': __maybe(__safeExp, 6, 2*5+3, )()}).evolver()"
    assert ''.join(list_item.parseString("{'a', 'b', 'c'}...")) == "__pset({'a', 'b', 'c'}).evolver()"
    assert ''.join(list_item.parseString("[2, 4, 6, 8, 10]...")) == "__pvector([2, 4, 6, 8, 10]).evolver()"


# TODO Stacking unary operators onto their respective positive/negative operands does not parse safely
#  line 105, in test_parse_arith
#     assert eval(stmt) == eval(''.join(arith_expr.parseString(stmt)))
# AssertionError: assert 1 == -1
#  +  where 1 = eval('--1-0//-1')
#  +  and   -1 = eval('__maybe(__safeFloorDiv, --1-0, -1)()')
#  +    where '__maybe(__safeFloorDiv, --1-0, -1)()' = <built-in method join of str object at 0x1033f0030>((['__maybe(__safeFloorDiv, --1-0, -1)()'], {}))
#  +      where <built-in method join of str object at 0x1033f0030> = ''.join
#  +      and   (['__maybe(__safeFloorDiv, --1-0, -1)()'], {}) = <bound method ParserElement.parseString of Combine:(Forward: bitwise operator term)>('--1-0//-1')
#  +        where <bound method ParserElement.parseString of Combine:(Forward: bitwise operator term)> = arith_expr.parseString
#
# Falsifying example: test_parse_arith(
#     p=1, q=0, binop='-', unop='-',
# )

# TODO Stacking division and floor division does not parse safely
#  line 120, in test_parse_arith
#     assert eval(stmt) == eval(''.join(arith_expr.parseString(stmt)))
# AssertionError: assert -1.0 == __undefined(None, (), {})
#  +  where -1.0 = eval('-1/1//1')
#  +  and   __undefined(None, (), {}) = eval('__maybe(__safeDiv, __maybe(__safeFloorDiv, -1, 1, 1)())()')
#  +    where '__maybe(__safeDiv, __maybe(__safeFloorDiv, -1, 1, 1)())()' = <built-in method join of str object at 0x10ac04030>((['__maybe(__safeDiv, __maybe(__safeFloorDiv, -1, 1, 1)())()'], {}))
#  +      where <built-in method join of str object at 0x10ac04030> = ''.join
#  +      and   (['__maybe(__safeDiv, __maybe(__safeFloorDiv, -1, 1, 1)())()'], {}) = <bound method ParserElement.parseString of Combine:(Forward: bitwise operator term)>('-1/1//1')
#  +        where <bound method ParserElement.parseString of Combine:(Forward: bitwise operator term)> = arith_expr.parseString
@hypothesis.settings(deadline=None)
@hypothesis.given(st.integers(min_value=1, max_value=255),  # exponent
                  st.integers(min_value=1),
                  st.text(['/', '*', '%', '+', '-'], min_size=1, max_size=1),
                  st.text(['+', '-'], min_size=1, max_size=1))
def test_parse_arith_expr(p, q, binop, unop):
    stmt = str(p) + binop + str(q)
    assert eval(stmt) == eval(''.join(list_item.parseString(stmt)))


@hypothesis.settings(deadline=None)
@hypothesis.given(st.integers(min_value=1, max_value=255),  # exponent
                  st.integers(min_value=1),
                  st.text(['/', '*', '%', '+', '-'], min_size=1, max_size=1),
                  st.text(['+', '-'], min_size=1, max_size=1),
                  st.from_regex(">=|<=|>|<|==", fullmatch=True))
def test_parse_comp_expr(p, q, binop, unop, compop):
    stmt = str(p) + binop + str(q) + ' ' + compop + ' ' + str(p) + binop + str(q)
    assert eval(stmt) == eval(''.join(list_item.parseString(stmt)))


bit_op = re.compile(r"<<|>>|&|\||\^")


@hypothesis.settings(deadline=None)
@hypothesis.given(st.integers(min_value=1, max_value=255),  # exponent
                  st.integers(min_value=1),
                  st.from_regex(bit_op, fullmatch=True))
def test_parse_bitwise(p, q, bitwise):
    assert str(p) + bitwise + str(q) == ''.join(list_item.parseString(str(p) + bitwise + str(q)))


# @pt.mark.filterwarnings('ignore::RuntimeWarning')
def test_compile_to_shared_object(inject_config):
    python_file_ = str(Path('examples/compiled.pyx'))
    full_path = Path('Aspidites/tests') / python_file_
    python_file = python_file_ if Path(woma_file).exists() else str(full_path)
    kwargs = get_cy_kwargs()
    code = setup_code(inject_config)
    kwargs.update(code=code, fname=python_file, bytecode=True, force=True, c=True, build_requires='', verbose=False)
    compile_args = CompilerArgs(**kwargs)
    try:
        Compiler(compile_args)
    except (FileNotFoundError, ValueError):
        kwargs.update(code=code, fname=Path(inject_config) / python_file_, bytecode=True)
        compile_args = CompilerArgs(**kwargs)
        Compiler(compile_args)

    from .examples.compiled import (Add, x, y, z, val, div_by_zero, Yield123, test_unary_sub, test_unary_add,
                                    test_factorial, match, nullity, D, idx)

    assert match() == 1
    assert match(0) == 1
    assert match(1) == 3
    assert match(4) == 7
    assert idx == 1

    with pt.raises(ContractNotRespected):
        # noinspection PyTypeChecker
        Add(x=6.5, y=12)

    assert [1, 2, 3] == [i for i in Yield123()]
    assert __maybe(Add, 6.5, 12)() == __undefined()
    assert x() == 6
    assert y() == __undefined()
    assert nullity == __undefined()
    assert z == 9
    assert Add(x=3, y=2) == 5
    assert val() == __undefined()
    assert div_by_zero == __undefined()

#
# def teardown_function():
#     for file in glob.glob('examples/compiled.*'):
#         os.remove(file)
#     try:
#         os.remove('examples/setup.py')
#     except FileNotFoundError:
#         pass



#!/usr/bin/python

#    Python/pyparsing educational microC compiler v1.0
#    Copyright (C) 2009  Zarko Zivanov
#    (largely based on flex/bison microC compiler by Zorica Suvajdzin, used with her permission;
#     current version can be found at http://www.acs.uns.ac.rs, under "Programski Prevodioci" [Serbian site])
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License can be found at <https://www.gnu.org/licenses/>.

from pyparsing import *
from sys import stdin, argv, exit

#defines debug level
# 0 - no debug
# 1 - print parsing results
# 2 - print parsing results and symbol table
# 3 - print parsing results only, without executing parse actions (grammar-only testing)
DEBUG = 0

##########################################################################################
##########################################################################################

#                               About microC language and microC compiler

# microC language and microC compiler are educational tools, and their goal is to show some basic principles
# of writing a C language compiler. Compiler represents one (relatively simple) solution, not necessarily the best one.
# This Python/pyparsing version is made using Python 2.6.4 and pyparsing 1.5.2 (and it may contain errors :) )

##########################################################################################
##########################################################################################

#                               Model of the used hypothetical processor

# The reason behind using a hypothetical processor is to simplify code generation and to concentrate on the compiler itself.
# This compiler can relatively easily be ported to x86, but one must know all the little details about which register
# can be used for what, which registers are default for various operations, etc.

# The hypothetical processor has 16 registers, called %0 to %15. Register %13 is used for the function return value (x86's eax),
# %14 is the stack frame pointer (x86's ebp) and %15 is the stack pointer (x86's esp). All data-handling instructions can be
# unsigned (suffix U), or signed (suffix S). These are ADD, SUB, MUL and DIV. These are three-address instructions,
# the first two operands are input, the third one is output. Whether these operands are registers, memory or constant
# is not relevant, all combinations are possible (except that output cannot be a constant). Constants are writen with a $ prefix (10-base only).
# Conditional jumps are handled by JXXY instructions, where XX is LT, GT, LE, GE, EQ, NE (less than, greater than, less than or equal, etc.)
# and Y is U or S (unsigned or signed, except for JEQ i JNE). Unconditional jump is JMP. The move instruction is MOV.
# Function handling is done using CALL, RET, PUSH and POP (C style function calls). Static data is defined using the WORD directive
# (example: variable: WORD 1), whose only argument defines the number of locations that are reserved.

##########################################################################################
##########################################################################################

#                               Grammar of The microC Programming Language
# (small subset of C made for compiler course at Faculty of Technical Sciences, Chair for Applied Computer Sciences, Novi Sad, Serbia)

# Patterns:

# letter
#  ->    "_" | "a" | "A" | "b" | "B" | "c" | "C" | "d" | "D" | "e" | "E" | "f"
#      | "F" | "g" | "G" | "h" | "H" | "i" | "I" | "j" | "J" | "k" | "K" | "l"
#      | "L" | "m" | "M" | "n" | "N" | "o" | "O" | "p" | "P" | "q" | "Q" | "r"
#      | "R" | "s" | "S" | "t" | "T" | "u" | "U" | "v" | "V" | "w" | "W" | "x"
#      | "X" | "y" | "Y" | "z" | "Z"

# digit
#  ->  "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"

# identifier
#  ->  letter ( letter | digit )*

# int_constant
#  ->  digit +

# unsigned_constant
#  ->  digit + ( "u" | "U" )

# Productions:

# program
#  ->  variable_list function_list
#  ->  function_list

# variable_list
#  ->  variable ";"
#  ->  variable_list variable ";"

# variable
#  ->  type identifier

# type
#  ->  "int"
#  ->  "unsigned"

# function_list
#  ->  function
#  ->  function_list function

# function
#  ->  type identifier "(" parameters ")" body

# parameters
#  ->  <empty>
#  ->  parameter_list

# parameter_list
#  ->  variable
#  ->  parameter_list "," variable

# body
#  ->  "{" variable_list statement_list "}"
#  ->  "{" statement_list "}"

# statement_list
#  ->  <empty>
#  ->  statement_list statement

# statement
#  ->  assignement_statement
#  ->  function_call_statement
#  ->  if_statement
#  ->  while_statement
#  ->  return_statement
#  ->  compound_statement

# assignement_statement
#  ->  identifier "=" num_exp ";"

# num_exp
#  ->  mul_exp
#  ->  num_exp "+" mul_exp
#  ->  num_exp "-" mul_exp

# mul_exp
#  ->  exp
#  ->  mul_exp "*" exp
#  ->  mul_exp "/" exp

# exp
#  ->  constant
#  ->  identifier
#  ->  function_call
#  ->  "(" num_exp ")"
#  ->  "+" exp
#  ->  "-" exp

# constant
#  ->  int_constant
#  ->  unsigned_constant

# function_call
#  ->  identifier "(" arguments ")"

# arguments
#  ->  <empty>
#  ->  argument_list

# argument_list
#  ->  num_exp
#  ->  argument_list "," num_exp

# function_call_statement
#  ->  function_call ";"

# if_statement
#  ->  "if" "(" log_exp ")" statement
#  ->  "if" "(" log_exp ")" statement "else" statement
#  ->   ->   ->   ->   ->   ->   -> ->   2

# log_exp
#  ->  and_exp
#  ->  log_exp "||" and_exp

# and_exp
#  ->  rel_exp
#  ->  and_exp "&&" rel_exp

# rel_exp
#  ->  num_exp  "<" num_exp
#  ->  num_exp  ">" num_exp
#  ->  num_exp  "<=" num_exp
#  ->  num_exp  ">=" num_exp
#  ->  num_exp  "==" num_exp
#  ->  num_exp  "!=" num_exp

# while_statement
#  ->  "while" "(" log_exp ")" statement

# return_statement
#  ->  "return" num_exp ";"

# compound_statement
#  ->  "{" statement_list "}"

# Comment: /* a comment */

##########################################################################################
##########################################################################################

class Enumerate(dict):
    """C enum emulation (original by Scott David Daniels)"""
    def __init__(self, names):
        for number, name in enumerate(names.split()):
            setattr(self, name, number)
            self[number] = name

class SharedData(object):
    """Data used in all three main classes"""

    #Possible kinds of symbol table entries
    KINDS = Enumerate("NO_KIND WORKING_REGISTER GLOBAL_VAR FUNCTION PARAMETER LOCAL_VAR CONSTANT")
    #Supported types of functions and variables
    TYPES = Enumerate("NO_TYPE INT UNSIGNED")

    #bit size of variables
    TYPE_BIT_SIZE = 16
    #min/max values of constants
    MIN_INT = -2 ** (TYPE_BIT_SIZE - 1)
    MAX_INT = 2 ** (TYPE_BIT_SIZE - 1) - 1
    MAX_UNSIGNED = 2 ** TYPE_BIT_SIZE - 1
    #available working registers (the last one is the register for function's return value!)
    REGISTERS = "%0 %1 %2 %3 %4 %5 %6 %7 %8 %9 %10 %11 %12 %13".split()
    #register for function's return value
    FUNCTION_REGISTER = len(REGISTERS) - 1
    #the index of last working register
    LAST_WORKING_REGISTER = len(REGISTERS) - 2
    #list of relational operators
    RELATIONAL_OPERATORS = "< > <= >= == !=".split()

    def __init__(self):
        #index of the currently parsed function
        self.functon_index = 0
        #name of the currently parsed function
        self.functon_name = 0
        #number of parameters of the currently parsed function
        self.function_params = 0
        #number of local variables of the currently parsed function
        self.function_vars = 0

##########################################################################################
##########################################################################################

class ExceptionSharedData(object):
    """Class for exception handling data"""

    def __init__(self):
        #position in currently parsed text
        self.location = 0
        #currently parsed text
        self.text = ""

    def setpos(self, location, text):
        """Helper function for setting curently parsed text and position"""
        self.location = location
        self.text = text

exshared = ExceptionSharedData()

class SemanticException(Exception):
    """Exception for semantic errors found during parsing, similar to ParseException.
       Introduced because ParseException is used internally in pyparsing and custom
       messages got lost and replaced by pyparsing's generic errors.
    """

    def __init__(self, message, print_location=True):
        super(SemanticException,self).__init__()
        self._message = message
        self.location = exshared.location
        self.print_location = print_location
        if exshared.location != None:
            self.line = lineno(exshared.location, exshared.text)
            self.col = col(exshared.location, exshared.text)
            self.text = line(exshared.location, exshared.text)
        else:
            self.line = self.col = self.text = None

    def _get_message(self):
        return self._message
    def _set_message(self, message):
        self._message = message
    message = property(_get_message, _set_message)

    def __str__(self):
        """String representation of the semantic error"""
        msg = "Error"
        if self.print_location and (self.line != None):
            msg += " at line %d, col %d" % (self.line, self.col)
        msg += ": %s" % self.message
        if self.print_location and (self.line != None):
            msg += "\n%s" % self.text
        return msg

##########################################################################################
##########################################################################################

class SymbolTableEntry(object):
    """Class which represents one symbol table entry."""

    def __init__(self, sname = "", skind = 0, stype = 0, sattr = None, sattr_name = "None"):
        """Initialization of symbol table entry.
           sname - symbol name
           skind - symbol kind
           stype - symbol type
           sattr - symbol attribute
           sattr_name - symbol attribute name (used only for table display)
        """
        self.name = sname
        self.kind = skind
        self.type = stype
        self.attribute = sattr
        self.attribute_name = sattr_name
        self.param_types = []

    def set_attribute(self, name, value):
        """Sets attribute's name and value"""
        self.attribute_name = name
        self.attribute = value

    def attribute_str(self):
        """Returns attribute string (used only for table display)"""
        return "{0}={1}".format(self.attribute_name, self.attribute) if self.attribute != None else "None"

class SymbolTable(object):
    """Class for symbol table of microC program"""

    def __init__(self, shared):
        """Initialization of the symbol table"""
        self.table = []
        self.lable_len = 0
        #put working registers in the symbol table
        for reg in range(SharedData.FUNCTION_REGISTER+1):
            self.insert_symbol(SharedData.REGISTERS[reg], SharedData.KINDS.WORKING_REGISTER, SharedData.TYPES.NO_TYPE)
        #shared data
        self.shared = shared

    def error(self, text=""):
        """Symbol table error exception. It should happen only if index is out of range while accessing symbol table.
           This exeption is not handled by the compiler, so as to allow traceback printing
        """
        if text == "":
            raise Exception("Symbol table index out of range")
        else:
            raise Exception("Symbol table error: %s" % text)

    def display(self):
        """Displays the symbol table content"""
        #Finding the maximum length for each column
        sym_name = "Symbol name"
        sym_len = max(max(len(i.name) for i in self.table),len(sym_name))
        kind_name = "Kind"
        kind_len = max(max(len(SharedData.KINDS[i.kind]) for i in self.table),len(kind_name))
        type_name = "Type"
        type_len = max(max(len(SharedData.TYPES[i.type]) for i in self.table),len(type_name))
        attr_name = "Attribute"
        attr_len = max(max(len(i.attribute_str()) for i in self.table),len(attr_name))
        #print table header
        print("{0:3s} | {1:^{2}s} | {3:^{4}s} | {5:^{6}s} | {7:^{8}} | {9:s}".format(" No", sym_name, sym_len, kind_name, kind_len, type_name, type_len, attr_name, attr_len, "Parameters"))
        print("-----------------------------" + "-" * (sym_len + kind_len + type_len + attr_len))
        #print symbol table
        for i,sym in enumerate(self.table):
            parameters = ""
            for p in sym.param_types:
                if parameters == "":
                    parameters = "{0}".format(SharedData.TYPES[p])
                else:
                    parameters += ", {0}".format(SharedData.TYPES[p])
            print("{0:3d} | {1:^{2}s} | {3:^{4}s} | {5:^{6}s} | {7:^{8}} | ({9})".format(i, sym.name, sym_len, SharedData.KINDS[sym.kind], kind_len, SharedData.TYPES[sym.type], type_len, sym.attribute_str(), attr_len, parameters))

    def insert_symbol(self, sname, skind, stype):
        """Inserts new symbol at the end of the symbol table.
           Returns symbol index
           sname - symbol name
           skind - symbol kind
           stype - symbol type
        """
        self.table.append(SymbolTableEntry(sname, skind, stype))
        self.table_len = len(self.table)
        return self.table_len-1

    def clear_symbols(self, index):
        """Clears all symbols begining with the index to the end of table"""
        try:
            del self.table[index:]
        except Exception:
            self.error()
        self.table_len = len(self.table)

    def lookup_symbol(self, sname, skind=list(SharedData.KINDS.keys()), stype=list(SharedData.TYPES.keys())):
        """Searches for symbol, from the end to the begining.
           Returns symbol index or None
           sname - symbol name
           skind - symbol kind (one kind, list of kinds, or None) deafult: any kind
           stype - symbol type (or None) default: any type
        """
        skind = skind if isinstance(skind, list) else [skind]
        stype = stype if isinstance(stype, list) else [stype]
        for i, sym in [[x, self.table[x]] for x in range(len(self.table) - 1, SharedData.LAST_WORKING_REGISTER, -1)]:
            if (sym.name == sname) and (sym.kind in skind) and (sym.type in stype):
                return i
        return None

    def insert_id(self, sname, skind, skinds, stype):
        """Inserts a new identifier at the end of the symbol table, if possible.
           Returns symbol index, or raises an exception if the symbol alredy exists
           sname   - symbol name
           skind   - symbol kind
           skinds  - symbol kinds to check for
           stype   - symbol type
        """
        index = self.lookup_symbol(sname, skinds)
        if index == None:
            index = self.insert_symbol(sname, skind, stype)
            return index
        else:
            raise SemanticException("Redefinition of '%s'" % sname)

    def insert_global_var(self, vname, vtype):
        "Inserts a new global variable"
        return self.insert_id(vname, SharedData.KINDS.GLOBAL_VAR, [SharedData.KINDS.GLOBAL_VAR, SharedData.KINDS.FUNCTION], vtype)

    def insert_local_var(self, vname, vtype, position):
        "Inserts a new local variable"
        index = self.insert_id(vname, SharedData.KINDS.LOCAL_VAR, [SharedData.KINDS.LOCAL_VAR, SharedData.KINDS.PARAMETER], vtype)
        self.table[index].attribute = position

    def insert_parameter(self, pname, ptype):
        "Inserts a new parameter"
        index = self.insert_id(pname, SharedData.KINDS.PARAMETER, SharedData.KINDS.PARAMETER, ptype)
        #set parameter's attribute to it's ordinal number
        self.table[index].set_attribute("Index", self.shared.function_params)
        #set parameter's type in param_types list of a function
        self.table[self.shared.function_index].param_types.append(ptype)
        return index

    def insert_function(self, fname, ftype):
        "Inserts a new function"
        index = self.insert_id(fname, SharedData.KINDS.FUNCTION, [SharedData.KINDS.GLOBAL_VAR, SharedData.KINDS.FUNCTION], ftype)
        self.table[index].set_attribute("Params",0)
        return index

    def insert_constant(self, cname, ctype):
        """Inserts a constant (or returns index if the constant already exists)
           Additionally, checks for range.
        """
        index = self.lookup_symbol(cname, stype=ctype)
        if index == None:
            num = int(cname)
            if ctype == SharedData.TYPES.INT:
                if (num < SharedData.MIN_INT) or (num > SharedData.MAX_INT):
                    raise SemanticException("Integer constant '%s' out of range" % cname)
            elif ctype == SharedData.TYPES.UNSIGNED:
                if (num < 0) or (num > SharedData.MAX_UNSIGNED):
                    raise SemanticException("Unsigned constant '%s' out of range" % cname)
            index = self.insert_symbol(cname, SharedData.KINDS.CONSTANT, ctype)
        return index

    def same_types(self, index1, index2):
        """Returns True if both symbol table elements are of the same type"""
        try:
            same = self.table[index1].type == self.table[index2].type != SharedData.TYPES.NO_TYPE
        except Exception:
            self.error()
        return same

    def same_type_as_argument(self, index, function_index, argument_number):
        """Returns True if index and function's argument are of the same type
           index - index in symbol table
           function_index - function's index in symbol table
           argument_number - # of function's argument
        """
        try:
            same = self.table[function_index].param_types[argument_number] == self.table[index].type
        except Exception:
            self.error()
        return same

    def get_attribute(self, index):
        try:
            return self.table[index].attribute
        except Exception:
            self.error()

    def set_attribute(self, index, value):
        try:
            self.table[index].attribute = value
        except Exception:
            self.error()

    def get_name(self, index):
        try:
            return self.table[index].name
        except Exception:
            self.error()

    def get_kind(self, index):
        try:
            return self.table[index].kind
        except Exception:
            self.error()

    def get_type(self, index):
        try:
            return self.table[index].type
        except Exception:
            self.error()

    def set_type(self, index, stype):
        try:
            self.table[index].type = stype
        except Exception:
            self.error()

##########################################################################################
##########################################################################################

class CodeGenerator(object):
    """Class for code generation methods."""

    #dictionary of relational operators
    RELATIONAL_DICT = {op:i for i, op in enumerate(SharedData.RELATIONAL_OPERATORS)}
    #conditional jumps for relational operators
    CONDITIONAL_JUMPS = ["JLTS", "JGTS", "JLES", "JGES", "JEQ ", "JNE ",
                         "JLTU", "JGTU", "JLEU", "JGEU", "JEQ ", "JNE "]
    #opposite conditional jumps for relational operators
    OPPOSITE_JUMPS = ["JGES", "JLES", "JGTS", "JLTS", "JNE ", "JEQ ",
                      "JGEU", "JLEU", "JGTU", "JLTU", "JNE ", "JEQ "]
    #supported operations
    OPERATIONS = {"+" : "ADD", "-" : "SUB", "*" : "MUL", "/" : "DIV"}
    #suffixes for signed and unsigned operations (if no type is specified, unsigned will be assumed)
    OPSIGNS = {SharedData.TYPES.NO_TYPE : "U", SharedData.TYPES.INT : "S", SharedData.TYPES.UNSIGNED : "U"}
    #text at start of data segment
    DATA_START_TEXT = "#DATA"
    #text at start of code segment
    CODE_START_TEXT = "#CODE"

    def __init__(self, shared, symtab):
        #generated code
        self.code = ""
        #prefix for internal labels
        self.internal = "@"
        #suffix for label definition
        self.definition = ":"
        #list of free working registers
        self.free_registers = list(range(SharedData.FUNCTION_REGISTER, -1, -1))
        #list of used working registers
        self.used_registers = []
        #list of used registers needed when function call is inside of a function call
        self.used_registers_stack = []
        #shared data
        self.shared = shared
        #symbol table
        self.symtab = symtab

    def error(self, text):
        """Compiler error exception. It should happen only if something is wrong with compiler.
           This exeption is not handled by the compiler, so as to allow traceback printing
        """
        raise Exception("Compiler error: %s" % text)

    def take_register(self, rtype = SharedData.TYPES.NO_TYPE):
        """Reserves one working register and sets its type"""
        if len(self.free_registers) == 0:
            self.error("no more free registers")
        reg = self.free_registers.pop()
        self.used_registers.append(reg)
        self.symtab.set_type(reg, rtype)
        return reg

    def take_function_register(self, rtype = SharedData.TYPES.NO_TYPE):
        """Reserves register for function return value and sets its type"""
        reg = SharedData.FUNCTION_REGISTER
        if reg not in self.free_registers:
            self.error("function register already taken")
        self.free_registers.remove(reg)
        self.used_registers.append(reg)
        self.symtab.set_type(reg, rtype)
        return reg

    def free_register(self, reg):
        """Releases working register"""
        if reg not in self.used_registers:
            self.error("register %s is not taken" % self.REGISTERS[reg])
        self.used_registers.remove(reg)
        self.free_registers.append(reg)
        self.free_registers.sort(reverse = True)

    def free_if_register(self, index):
        """If index is a working register, free it, otherwise just return (helper function)"""
        if (index < 0) or (index > SharedData.FUNCTION_REGISTER):
            return
        else:
            self.free_register(index)

    def label(self, name, internal=False, definition=False):
        """Generates label name (helper function)
           name - label name
           internal - boolean value, adds "@" prefix to label
           definition - boolean value, adds ":" suffix to label
        """
        return "{0}{1}{2}".format(self.internal if internal else "", name, self.definition if definition else "")

    def symbol(self, index):
        """Generates symbol name from index"""
        #if index is actually a string, just return it
        if isinstance(index, str):
            return index
        elif (index < 0) or (index >= self.symtab.table_len):
            self.error("symbol table index out of range")
        sym = self.symtab.table[index]
        #local variables are located at negative offset from frame pointer register
        if sym.kind == SharedData.KINDS.LOCAL_VAR:
            return "-{0}(1:%14)".format(sym.attribute * 4 + 4)
        #parameters are located at positive offset from frame pointer register
        elif sym.kind == SharedData.KINDS.PARAMETER:
            return "{0}(1:%14)".format(8 + sym.attribute * 4)
        elif sym.kind == SharedData.KINDS.CONSTANT:
            return "${0}".format(sym.name)
        else:
            return "{0}".format(sym.name)

    def save_used_registers(self):
        """Pushes all used working registers before function call"""
        used = self.used_registers[:]
        del self.used_registers[:]
        self.used_registers_stack.append(used[:])
        used.sort()
        for reg in used:
            self.newline_text("PUSH\t%s" % SharedData.REGISTERS[reg], True)
        self.free_registers.extend(used)
        self.free_registers.sort(reverse = True)

    def restore_used_registers(self):
        """Pops all used working registers after function call"""
        used = self.used_registers_stack.pop()
        self.used_registers = used[:]
        used.sort(reverse = True)
        for reg in used:
            self.newline_text("POP \t%s" % SharedData.REGISTERS[reg], True)
            self.free_registers.remove(reg)

    def text(self, text):
        """Inserts text into generated code"""
        self.code += text

    def prepare_data_segment(self):
        """Inserts text at the start of data segment"""
        self.text(self.DATA_START_TEXT)

    def prepare_code_segment(self):
        """Inserts text at the start of code segment"""
        self.newline_text(self.CODE_START_TEXT)

    def newline(self, indent=False):
        """Inserts a newline, optionally with indentation."""
        self.text("\n")
        if indent:
            self.text("\t\t\t")

    def newline_text(self, text, indent = False):
        """Inserts a newline and text, optionally with indentation (helper function)"""
        self.newline(indent)
        self.text(text)

    def newline_label(self, name, internal=False, definition=False):
        """Inserts a newline and a label (helper function)
           name - label name
           internal - boolean value, adds "@" prefix to label
           definition - boolean value, adds ":" suffix to label
        """
        self.newline_text(self.label("{0}{1}{2}".format("@" if internal else "", name, ":" if definition else "")))

    def global_var(self, name):
        """Inserts a new static (global) variable definition"""
        self.newline_label(name, False, True)
        self.newline_text("WORD\t1", True)

    def arithmetic_mnemonic(self, op_name, op_type):
        """Generates an arithmetic instruction mnemonic"""
        return self.OPERATIONS[op_name] + self.OPSIGNS[op_type]

    def arithmetic(self, operation, operand1, operand2, operand3 = None):
        """Generates an arithmetic instruction
           operation - one of supporetd operations
           operandX - index in symbol table or text representation of operand
           First two operands are input, third one is output
        """
        if isinstance(operand1, int):
            output_type = self.symtab.get_type(operand1)
            self.free_if_register(operand1)
        else:
            output_type = None
        if isinstance(operand2, int):
            output_type = self.symtab.get_type(operand2) if output_type == None else output_type
            self.free_if_register(operand2)
        else:
            output_type = SharedData.TYPES.NO_TYPE if output_type == None else output_type
        #if operand3 is not defined, reserve one free register for it
        output = self.take_register(output_type) if operand3 == None else operand3
        mnemonic = self.arithmetic_mnemonic(operation, output_type)
        self.newline_text("{0}\t{1},{2},{3}".format(mnemonic, self.symbol(operand1), self.symbol(operand2), self.symbol(output)), True)
        return output

    def relop_code(self, relop, operands_type):
        """Returns code for relational operator
           relop - relational operator
           operands_type - int or unsigned
        """
        code = self.RELATIONAL_DICT[relop]
        offset = 0 if operands_type == SharedData.TYPES.INT else len(SharedData.RELATIONAL_OPERATORS)
        return code + offset

    def jump(self, relcode, opposite, label):
        """Generates a jump instruction
           relcode  - relational operator code
           opposite - generate normal or opposite jump
           label    - jump label
        """
        jump = self.OPPOSITE_JUMPS[relcode] if opposite else self.CONDITIONAL_JUMPS[relcode]
        self.newline_text("{0}\t{1}".format(jump, label), True)

    def unconditional_jump(self, label):
        """Generates an unconditional jump instruction
           label    - jump label
        """
        self.newline_text("JMP \t{0}".format(label), True)

    def move(self,operand1, operand2):
        """Generates a move instruction
           If the output operand (opernad2) is a working register, sets it's type
           operandX - index in symbol table or text representation of operand
        """
        if isinstance(operand1, int):
            output_type = self.symtab.get_type(operand1)
            self.free_if_register(operand1)
        else:
            output_type = SharedData.TYPES.NO_TYPE
        self.newline_text("MOV \t{0},{1}".format(self.symbol(operand1), self.symbol(operand2)), True)
        if isinstance(operand2, int):
            if self.symtab.get_kind(operand2) == SharedData.KINDS.WORKING_REGISTER:
                self.symtab.set_type(operand2, output_type)

    def push(self, operand):
        """Generates a push operation"""
        self.newline_text("PUSH\t%s" % self.symbol(operand), True)

    def pop(self, operand):
        """Generates a pop instruction"""
        self.newline_text("POP \t%s" % self.symbol(operand), True)

    def compare(self, operand1, operand2):
        """Generates a compare instruction
           operandX - index in symbol table
        """
        typ = self.symtab.get_type(operand1)
        self.free_if_register(operand1)
        self.free_if_register(operand2)
        self.newline_text("CMP{0}\t{1},{2}".format(self.OPSIGNS[typ], self.symbol(operand1), self.symbol(operand2)), True)

    def function_begin(self):
        """Inserts function name label and function frame initialization"""
        self.newline_label(self.shared.function_name, False, True)
        self.push("%14")
        self.move("%15", "%14")

    def function_body(self):
        """Inserts a local variable initialization and body label"""
        if self.shared.function_vars > 0:
            const = self.symtab.insert_constant("0{}".format(self.shared.function_vars * 4), SharedData.TYPES.UNSIGNED)
            self.arithmetic("-", "%15", const, "%15")
        self.newline_label(self.shared.function_name + "_body", True, True)

    def function_end(self):
        """Inserts an exit label and function return instructions"""
        self.newline_label(self.shared.function_name + "_exit", True, True)
        self.move("%14", "%15")
        self.pop("%14")
        self.newline_text("RET", True)

    def function_call(self, function, arguments):
        """Generates code for a function call
           function - function index in symbol table
           arguments - list of arguments (indexes in symbol table)
        """
        #push each argument to stack
        for arg in arguments:
            self.push(self.symbol(arg))
            self.free_if_register(arg)
        self.newline_text("CALL\t"+self.symtab.get_name(function), True)
        args = self.symtab.get_attribute(function)
        #generates stack cleanup if function has arguments
        if args > 0:
            args_space = self.symtab.insert_constant("{0}".format(args * 4), SharedData.TYPES.UNSIGNED)
            self.arithmetic("+", "%15", args_space, "%15")

##########################################################################################
##########################################################################################

class MicroC(object):
    """Class for microC parser/compiler"""

    def __init__(self):
        #Definitions of terminal symbols for microC programming language
        self.tId = Word(alphas+"_",alphanums+"_")
        self.tInteger = Word(nums).setParseAction(lambda x : [x[0], SharedData.TYPES.INT])
        self.tUnsigned = Regex(r"[0-9]+[uU]").setParseAction(lambda x : [x[0][:-1], SharedData.TYPES.UNSIGNED])
        self.tConstant = (self.tUnsigned | self.tInteger).setParseAction(self.constant_action)
        self.tType = Keyword("int").setParseAction(lambda x : SharedData.TYPES.INT) | \
                     Keyword("unsigned").setParseAction(lambda x : SharedData.TYPES.UNSIGNED)
        self.tRelOp = oneOf(SharedData.RELATIONAL_OPERATORS)
        self.tMulOp = oneOf("* /")
        self.tAddOp = oneOf("+ -")

        #Definitions of rules for global variables
        self.rGlobalVariable = (self.tType("type") + self.tId("name") +
                                FollowedBy(";")).setParseAction(self.global_variable_action)
        self.rGlobalVariableList = ZeroOrMore(self.rGlobalVariable + Suppress(";"))

        #Definitions of rules for numeric expressions
        self.rExp = Forward()
        self.rMulExp = Forward()
        self.rNumExp = Forward()
        self.rArguments = delimitedList(self.rNumExp("exp").setParseAction(self.argument_action))
        self.rFunctionCall = ((self.tId("name") + FollowedBy("(")).setParseAction(self.function_call_prepare_action) +
                              Suppress("(") + Optional(self.rArguments)("args") + Suppress(")")).setParseAction(self.function_call_action)
        self.rExp << (self.rFunctionCall |
                      self.tConstant |
                      self.tId("name").setParseAction(self.lookup_id_action) |
                      Group(Suppress("(") + self.rNumExp + Suppress(")")) |
                      Group("+" + self.rExp) |
                      Group("-" + self.rExp)).setParseAction(lambda x : x[0])
        self.rMulExp << ((self.rExp + ZeroOrMore(self.tMulOp + self.rExp))).setParseAction(self.mulexp_action)
        self.rNumExp << (self.rMulExp + ZeroOrMore(self.tAddOp + self.rMulExp)).setParseAction(self.numexp_action)

        #Definitions of rules for logical expressions (these are without parenthesis support)
        self.rAndExp = Forward()
        self.rLogExp = Forward()
        self.rRelExp = (self.rNumExp + self.tRelOp + self.rNumExp).setParseAction(self.relexp_action)
        self.rAndExp << (self.rRelExp("exp") + ZeroOrMore(Literal("&&").setParseAction(self.andexp_action) +
                         self.rRelExp("exp")).setParseAction(lambda x : self.relexp_code))
        self.rLogExp << (self.rAndExp("exp") + ZeroOrMore(Literal("||").setParseAction(self.logexp_action) +
                         self.rAndExp("exp")).setParseAction(lambda x : self.andexp_code))

        #Definitions of rules for statements
        self.rStatement = Forward()
        self.rStatementList = Forward()
        self.rReturnStatement = (Keyword("return") + self.rNumExp("exp") +
                                 Suppress(";")).setParseAction(self.return_action)
        self.rAssignmentStatement = (self.tId("var") + Suppress("=") + self.rNumExp("exp") +
                                     Suppress(";")).setParseAction(self.assignment_action)
        self.rFunctionCallStatement = self.rFunctionCall + Suppress(";")
        self.rIfStatement = ( (Keyword("if") + FollowedBy("(")).setParseAction(self.if_begin_action) +
                              (Suppress("(") + self.rLogExp + Suppress(")")).setParseAction(self.if_body_action) +
                              (self.rStatement + Empty()).setParseAction(self.if_else_action) +
                              Optional(Keyword("else") + self.rStatement)).setParseAction(self.if_end_action)
        self.rWhileStatement = ( (Keyword("while") + FollowedBy("(")).setParseAction(self.while_begin_action) +
                                 (Suppress("(") + self.rLogExp + Suppress(")")).setParseAction(self.while_body_action) +
                                 self.rStatement).setParseAction(self.while_end_action)
        self.rCompoundStatement = Group(Suppress("{") + self.rStatementList + Suppress("}"))
        self.rStatement << (self.rReturnStatement | self.rIfStatement | self.rWhileStatement |
                            self.rFunctionCallStatement | self.rAssignmentStatement | self.rCompoundStatement)
        self.rStatementList << ZeroOrMore(self.rStatement)

        self.rLocalVariable = (self.tType("type") + self.tId("name") + FollowedBy(";")).setParseAction(self.local_variable_action)
        self.rLocalVariableList = ZeroOrMore(self.rLocalVariable + Suppress(";"))
        self.rFunctionBody = Suppress("{") + Optional(self.rLocalVariableList).setParseAction(self.function_body_action) + \
                             self.rStatementList + Suppress("}")
        self.rParameter = (self.tType("type") + self.tId("name")).setParseAction(self.parameter_action)
        self.rParameterList = delimitedList(self.rParameter)
        self.rFunction = ( (self.tType("type") + self.tId("name")).setParseAction(self.function_begin_action) +
                           Group(Suppress("(") + Optional(self.rParameterList)("params") + Suppress(")") +
                           self.rFunctionBody)).setParseAction(self.function_end_action)

        self.rFunctionList = OneOrMore(self.rFunction)
        self.rProgram = (Empty().setParseAction(self.data_begin_action) + self.rGlobalVariableList +
                         Empty().setParseAction(self.code_begin_action) + self.rFunctionList).setParseAction(self.program_end_action)

        #shared data
        self.shared = SharedData()
        #symbol table
        self.symtab = SymbolTable(self.shared)
        #code generator
        self.codegen = CodeGenerator(self.shared, self.symtab)

        #index of the current function call
        self.function_call_index = -1
        #stack for the nested function calls
        self.function_call_stack = []
        #arguments of the current function call
        self.function_arguments = []
        #stack for arguments of the nested function calls
        self.function_arguments_stack = []
        #number of arguments for the curent function call
        self.function_arguments_number = -1
        #stack for the number of arguments for the nested function calls
        self.function_arguments_number_stack = []

        #last relational expression
        self.relexp_code = None
        #last and expression
        self.andexp_code = None
        #label number for "false" internal labels
        self.false_label_number = -1
        #label number for all other internal labels
        self.label_number = None
        #label stack for nested statements
        self.label_stack = []

    def warning(self, message, print_location=True):
        """Displays warning message. Uses exshared for current location of parsing"""
        msg = "Warning"
        if print_location and (exshared.location != None):
            wline = lineno(exshared.location, exshared.text)
            wcol = col(exshared.location, exshared.text)
            wtext = line(exshared.location, exshared.text)
            msg += " at line %d, col %d" % (wline, wcol)
        msg += ": %s" % message
        if print_location and (exshared.location != None):
            msg += "\n%s" % wtext
        print(msg)


    def data_begin_action(self):
        """Inserts text at start of data segment"""
        self.codegen.prepare_data_segment()

    def code_begin_action(self):
        """Inserts text at start of code segment"""
        self.codegen.prepare_code_segment()

    def global_variable_action(self, text, loc, var):
        """Code executed after recognising a global variable"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("GLOBAL_VAR:",var)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        index = self.symtab.insert_global_var(var.name, var.type)
        self.codegen.global_var(var.name)
        return index

    def local_variable_action(self, text, loc, var):
        """Code executed after recognising a local variable"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("LOCAL_VAR:",var, var.name, var.type)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        index = self.symtab.insert_local_var(var.name, var.type, self.shared.function_vars)
        self.shared.function_vars += 1
        return index

    def parameter_action(self, text, loc, par):
        """Code executed after recognising a parameter"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("PARAM:",par)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        index = self.symtab.insert_parameter(par.name, par.type)
        self.shared.function_params += 1
        return index

    def constant_action(self, text, loc, const):
        """Code executed after recognising a constant"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("CONST:",const)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        return self.symtab.insert_constant(const[0], const[1])

    def function_begin_action(self, text, loc, fun):
        """Code executed after recognising a function definition (type and function name)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("FUN_BEGIN:",fun)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        self.shared.function_index = self.symtab.insert_function(fun.name, fun.type)
        self.shared.function_name = fun.name
        self.shared.function_params = 0
        self.shared.function_vars = 0
        self.codegen.function_begin();

    def function_body_action(self, text, loc, fun):
        """Code executed after recognising the beginning of function's body"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("FUN_BODY:",fun)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        self.codegen.function_body()

    def function_end_action(self, text, loc, fun):
        """Code executed at the end of function definition"""
        if DEBUG > 0:
            print("FUN_END:",fun)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #set function's attribute to number of function parameters
        self.symtab.set_attribute(self.shared.function_index, self.shared.function_params)
        #clear local function symbols (but leave function name)
        self.symtab.clear_symbols(self.shared.function_index + 1)
        self.codegen.function_end()

    def return_action(self, text, loc, ret):
        """Code executed after recognising a return statement"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("RETURN:",ret)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        if not self.symtab.same_types(self.shared.function_index, ret.exp[0]):
            raise SemanticException("Incompatible type in return")
        #set register for function's return value to expression value
        reg = self.codegen.take_function_register()
        self.codegen.move(ret.exp[0], reg)
        #after return statement, register for function's return value is available again
        self.codegen.free_register(reg)
        #jump to function's exit
        self.codegen.unconditional_jump(self.codegen.label(self.shared.function_name+"_exit", True))

    def lookup_id_action(self, text, loc, var):
        """Code executed after recognising an identificator in expression"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("EXP_VAR:",var)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        var_index = self.symtab.lookup_symbol(var.name, [SharedData.KINDS.GLOBAL_VAR, SharedData.KINDS.PARAMETER, SharedData.KINDS.LOCAL_VAR])
        if var_index == None:
            raise SemanticException("'%s' undefined" % var.name)
        return var_index

    def assignment_action(self, text, loc, assign):
        """Code executed after recognising an assignment statement"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("ASSIGN:",assign)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        var_index = self.symtab.lookup_symbol(assign.var, [SharedData.KINDS.GLOBAL_VAR, SharedData.KINDS.PARAMETER, SharedData.KINDS.LOCAL_VAR])
        if var_index == None:
            raise SemanticException("Undefined lvalue '%s' in assignment" % assign.var)
        if not self.symtab.same_types(var_index, assign.exp[0]):
            raise SemanticException("Incompatible types in assignment")
        self.codegen.move(assign.exp[0], var_index)

    def mulexp_action(self, text, loc, mul):
        """Code executed after recognising a mulexp expression (something *|/ something)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("MUL_EXP:",mul)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #iterate through all multiplications/divisions
        m = list(mul)
        while len(m) > 1:
            if not self.symtab.same_types(m[0], m[2]):
                raise SemanticException("Invalid opernads to binary '%s'" % m[1])
            reg = self.codegen.arithmetic(m[1], m[0], m[2])
            #replace first calculation with it's result
            m[0:3] = [reg]
        return m[0]

    def numexp_action(self, text, loc, num):
        """Code executed after recognising a numexp expression (something +|- something)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("NUM_EXP:",num)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #iterate through all additions/substractions
        n = list(num)
        while len(n) > 1:
            if not self.symtab.same_types(n[0], n[2]):
                raise SemanticException("Invalid opernads to binary '%s'" % n[1])
            reg = self.codegen.arithmetic(n[1], n[0], n[2])
            #replace first calculation with it's result
            n[0:3] = [reg]
        return n[0]

    def function_call_prepare_action(self, text, loc, fun):
        """Code executed after recognising a function call (type and function name)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("FUN_PREP:",fun)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        index = self.symtab.lookup_symbol(fun.name, SharedData.KINDS.FUNCTION)
        if index == None:
            raise SemanticException("'%s' is not a function" % fun.name)
        #save any previous function call data (for nested function calls)
        self.function_call_stack.append(self.function_call_index)
        self.function_call_index = index
        self.function_arguments_stack.append(self.function_arguments[:])
        del self.function_arguments[:]
        self.codegen.save_used_registers()

    def argument_action(self, text, loc, arg):
        """Code executed after recognising each of function's arguments"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("ARGUMENT:",arg.exp)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        arg_ordinal = len(self.function_arguments)
        #check argument's type
        if not self.symtab.same_type_as_argument(arg.exp, self.function_call_index, arg_ordinal):
            raise SemanticException("Incompatible type for argument %d in '%s'" % (arg_ordinal + 1, self.symtab.get_name(self.function_call_index)))
        self.function_arguments.append(arg.exp)

    def function_call_action(self, text, loc, fun):
        """Code executed after recognising the whole function call"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("FUN_CALL:",fun)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #check number of arguments
        if len(self.function_arguments) != self.symtab.get_attribute(self.function_call_index):
            raise SemanticException("Wrong number of arguments for function '%s'" % fun.name)
        #arguments should be pushed to stack in reverse order
        self.function_arguments.reverse()
        self.codegen.function_call(self.function_call_index, self.function_arguments)
        self.codegen.restore_used_registers()
        return_type = self.symtab.get_type(self.function_call_index)
        #restore previous function call data
        self.function_call_index = self.function_call_stack.pop()
        self.function_arguments = self.function_arguments_stack.pop()
        register = self.codegen.take_register(return_type)
        #move result to a new free register, to allow the next function call
        self.codegen.move(self.codegen.take_function_register(return_type), register)
        return register

    def relexp_action(self, text, loc, arg):
        """Code executed after recognising a relexp expression (something relop something)"""
        if DEBUG > 0:
            print("REL_EXP:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        exshared.setpos(loc, text)
        if not self.symtab.same_types(arg[0], arg[2]):
            raise SemanticException("Invalid operands for operator '{0}'".format(arg[1]))
        self.codegen.compare(arg[0], arg[2])
        #return relational operator's code
        self.relexp_code = self.codegen.relop_code(arg[1], self.symtab.get_type(arg[0]))
        return self.relexp_code

    def andexp_action(self, text, loc, arg):
        """Code executed after recognising a andexp expression (something and something)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("AND+EXP:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        label = self.codegen.label("false{0}".format(self.false_label_number), True, False)
        self.codegen.jump(self.relexp_code, True, label)
        self.andexp_code = self.relexp_code
        return self.andexp_code

    def logexp_action(self, text, loc, arg):
        """Code executed after recognising logexp expression (something or something)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("LOG_EXP:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        label = self.codegen.label("true{0}".format(self.label_number), True, False)
        self.codegen.jump(self.relexp_code, False, label)
        self.codegen.newline_label("false{0}".format(self.false_label_number), True, True)
        self.false_label_number += 1

    def if_begin_action(self, text, loc, arg):
        """Code executed after recognising an if statement (if keyword)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("IF_BEGIN:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        self.false_label_number += 1
        self.label_number = self.false_label_number
        self.codegen.newline_label("if{0}".format(self.label_number), True, True)

    def if_body_action(self, text, loc, arg):
        """Code executed after recognising if statement's body"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("IF_BODY:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #generate conditional jump (based on last compare)
        label = self.codegen.label("false{0}".format(self.false_label_number), True, False)
        self.codegen.jump(self.relexp_code, True, label)
        #generate 'true' label (executes if condition is satisfied)
        self.codegen.newline_label("true{0}".format(self.label_number), True, True)
        #save label numbers (needed for nested if/while statements)
        self.label_stack.append(self.false_label_number)
        self.label_stack.append(self.label_number)

    def if_else_action(self, text, loc, arg):
        """Code executed after recognising if statement's else body"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("IF_ELSE:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #jump to exit after all statements for true condition are executed
        self.label_number = self.label_stack.pop()
        label = self.codegen.label("exit{0}".format(self.label_number), True, False)
        self.codegen.unconditional_jump(label)
        #generate final 'false' label (executes if condition isn't satisfied)
        self.codegen.newline_label("false{0}".format(self.label_stack.pop()), True, True)
        self.label_stack.append(self.label_number)

    def if_end_action(self, text, loc, arg):
        """Code executed after recognising a whole if statement"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("IF_END:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        self.codegen.newline_label("exit{0}".format(self.label_stack.pop()), True, True)

    def while_begin_action(self, text, loc, arg):
        """Code executed after recognising a while statement (while keyword)"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("WHILE_BEGIN:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        self.false_label_number += 1
        self.label_number = self.false_label_number
        self.codegen.newline_label("while{0}".format(self.label_number), True, True)

    def while_body_action(self, text, loc, arg):
        """Code executed after recognising while statement's body"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("WHILE_BODY:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #generate conditional jump (based on last compare)
        label = self.codegen.label("false{0}".format(self.false_label_number), True, False)
        self.codegen.jump(self.relexp_code, True, label)
        #generate 'true' label (executes if condition is satisfied)
        self.codegen.newline_label("true{0}".format(self.label_number), True, True)
        self.label_stack.append(self.false_label_number)
        self.label_stack.append(self.label_number)

    def while_end_action(self, text, loc, arg):
        """Code executed after recognising a whole while statement"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("WHILE_END:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        #jump to condition checking after while statement body
        self.label_number = self.label_stack.pop()
        label = self.codegen.label("while{0}".format(self.label_number), True, False)
        self.codegen.unconditional_jump(label)
        #generate final 'false' label and exit label
        self.codegen.newline_label("false{0}".format(self.label_stack.pop()), True, True)
        self.codegen.newline_label("exit{0}".format(self.label_number), True, True)

    def program_end_action(self, text, loc, arg):
        """Checks if there is a 'main' function and the type of 'main' function"""
        exshared.setpos(loc, text)
        if DEBUG > 0:
            print("PROGRAM_END:",arg)
            if DEBUG == 2: self.symtab.display()
            if DEBUG > 2: return
        index = self.symtab.lookup_symbol("main",SharedData.KINDS.FUNCTION)
        if index == None:
            raise SemanticException("Undefined reference to 'main'", False)
        elif self.symtab.get_type(index) != SharedData.TYPES.INT:
            self.warning("Return type of 'main' is not int", False)

    def parse_text(self,text):
        """Parse string (helper function)"""
        try:
            return self.rProgram.ignore(cStyleComment).parseString(text, parseAll=True)
        except SemanticException as err:
            print(err)
            exit(3)
        except ParseException as err:
            print(err)
            exit(3)

    def parse_file(self,filename):
        """Parse file (helper function)"""
        try:
            return self.rProgram.ignore(cStyleComment).parseFile(filename, parseAll=True)
        except SemanticException as err:
            print(err)
            exit(3)
        except ParseException as err:
            print(err)
            exit(3)

##########################################################################################
##########################################################################################
if 0:
    #main program
    mc = MicroC()
    output_file = "output.asm"

    if len(argv) == 1:
        input_file = stdin
    elif len(argv) == 2:
        input_file = argv[1]
    elif len(argv) == 3:
        input_file = argv[1]
        output_file = argv[2]
    else:
        usage = """Usage: {0} [input_file [output_file]]
    If output file is omitted, output.asm is used
    If input file is omitted, stdin is used""".format(argv[0])
        print(usage)
        exit(1)
    try:
        parse = stdin if input_file == stdin else open(input_file,'r')
    except Exception:
        print("Input file '%s' open error" % input_file)
        exit(2)
    mc.parse_file(parse)
    #if you want to see the final symbol table, uncomment next line
    #mc.symtab.display()
    try:
        out = open(output_file, 'w')
        out.write(mc.codegen.code)
        out.close
    except Exception:
        print("Output file '%s' open error" % output_file)
        exit(2)

##########################################################################################
##########################################################################################

if __name__ == "__main__":

    test_program_example = """
        int a;
        int b;
        int c;
        unsigned d;

        int fun1(int x, unsigned y) {
            return 123;
        }

        int fun2(int a) {
            return 1 + a * fun1(a, 456u);
        }

        int main(int x, int y) {
            int w;
            unsigned z;
            if (9 > 8 && 2 < 3 || 6 != 5 && a <= b && c < x || w >= y) {
                a = b + 1;
                if (x == y)
                    while (d < 4u)
                        x = x * w;
                else
                    while (a + b < c - y && x > 3 || y < 2)
                        if (z > d)
                            a = a - 4;
                        else
                            b = a * b * c * x / y;
            }
            else
                c = 4;
            a = fun1(x,d) + fun2(fun1(fun2(w + 3 * 2) + 2 * c, 2u));
            return 2;
        }
    """

    mc = MicroC()
    mc.parse_text(test_program_example)
    print(mc.codegen.code)



# getNTPserversNew.py
#
# Demonstration of the parsing module, implementing a HTML page scanner,
# to extract a list of NTP time servers from the NIST web site.
#
# Copyright 2004-2010, by Paul McGuire
# September, 2010 - updated to more current use of setResultsName, new NIST URL
#
from ....._vendor import pyparsing as pp
ppc = pp.pyparsing_common
from contextlib import closing

try:
    import urllib.request
    urlopen = urllib.request.urlopen
except ImportError:
    import urllib
    urlopen = urllib.urlopen

integer = pp.Word(pp.nums)
ipAddress = ppc.ipv4_address()
hostname = pp.delimitedList(pp.Word(pp.alphas, pp.alphanums+"-_"), ".", combine=True)
tdStart, tdEnd = pp.makeHTMLTags("td")
timeServerPattern = (tdStart + hostname("hostname") + tdEnd
                     + tdStart + ipAddress("ipAddr") + tdEnd
                     + tdStart + tdStart.tag_body("loc") + tdEnd)

# get list of time servers
nistTimeServerURL = "https://tf.nist.gov/tf-cgi/servers.cgi#"
with closing(urlopen(nistTimeServerURL)) as serverListPage:
    serverListHTML = serverListPage.read().decode("UTF-8")

addrs = {}
for srvr, startloc, endloc in timeServerPattern.scanString(serverListHTML):
    print("{0} ({1}) - {2}".format(srvr.ipAddr, srvr.hostname.strip(), srvr.loc.strip()))
    addrs[srvr.ipAddr] = srvr.loc



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from Aspidites._vendor.pyparsing import oneOf, Keyword, replaceWith, Literal, Suppress, Group, Regex

available_bool_pragmas = [
    'cython.binding',
    'cython.boundscheck',
    'cython.wraparound',
    'cython.initializedcheck',
    'cython.nonecheck',
    'cython.overflowcheck',
    'cython.overflowcheck.fold',
    'cython.embedsignature',
    'cython.cdivision',
    'cython.cdivision_warnings',
    'cython.always_allow_keywords',
    'cython.c_api_binop_methods',
    'cython.profile',
    'cython.linetrace',
    'cython.infer_types',
    'cython.type_version_tag',
    'cython.unraisable_tracebacks',
    'cython.iterable_coroutine',
    'cython.emit_code_comments',
    'cython.cpp_locals',
    'cython.optimize.use_switch',
    'cython.optimize.unpack_method_calls',
    'cython.warn.undeclared',
    'cython.warn.unreachable',
    'cython.warn.maybe_uninitialized',
    'cython.warn.unused',
    'cython.warn.unused_arg',
    'cython.warn.unused_result',
    'cython.warn.multiple_declarators'
]
available_pragmas = [
    'cfunc',
    'ccall',
    'nogil',
    'no_gc',
    'inline',
    'curried'
]
indent = '    '
nl_indent = '\n' + indent
sep = ", "
pragma = Literal("#").setParseAction(replaceWith('@')).setName('pragma')
private_def_decl = Literal("(").setParseAction(replaceWith("def ")).setName("function declarator")
args_end = Group(Literal(")") + Literal(")")).setParseAction(replaceWith(") -> ")).setName("return annotation")
underscore = Literal("_").setName("lambda variable")
lparen, rparen, lbrack, rbrack, lbrace, rbrace, colon, comma = map(Suppress, "()[]{}:,")
lit_lparen, lit_rparen, lit_lbrack, lit_rbrack, lit_lbrace, lit_rbrace, lit_colon, lit_comma = map(str, "()[]{}:,")
bool_literal = oneOf("True False", asKeyword=True).setParseAction(lambda s, l, t: t[0] == "True").setName('boolean literal')
nullit = oneOf("/0 ø Ø").setParseAction(replaceWith("__undefined()")).setName('nullity')
signop = Literal("+") | Literal("-")
multop = Literal("*") | oneOf("/ //") | Literal("%") | Literal("@")
plusop = Literal("+") | Literal("-")
exponp = Literal("**")
factop = Literal("!")
bitwiseop = Regex(r"<<|>>|&|\||\^").setName("bitwise operator")
andop = Literal('&&').setParseAction(replaceWith(' and '))
orop = Literal('||').setParseAction(replaceWith(' or '))
comparisonop = andop | orop | Regex(">=|<=|>|<|==").setName("comparison operator")
assign_eq = Literal("=").setName('assignment')
noclosure = Literal("...").setName('ellipsis')  # That's no closure
index_items = Literal("[!]").setName('index trigram')
keys = Literal("[@]").setName('keys trigram')
persist = Literal("[.]").setParseAction(replaceWith('.persistent()')).setName('persist trigram')
count_vals = Literal("[#]").setName('count trigram')
set_add = Literal("[$]").setName('set trigram')
discard = Literal("[%]").setName('discard trigram')
copy = Literal("[&]").setName('copy trigram')
append = Literal("[^]").setName('copy trigram')
update = Literal("[+]").setName('update trigram')
remove = Literal("[*]").setName('remove trigram')
match_none = Literal("(!)").setParseAction(replaceWith('__match')).setName('match trigram')
context_stmt = Literal("<!>").setParseAction(replaceWith("with ")).setName('context trigram')
return_none = Literal("<*>").setParseAction(replaceWith("return ")).setName('return trigram')
yield_none = Literal("<^>").setParseAction(replaceWith("yield ")).setName('yield trigram')
for_none = Literal("<@>").setParseAction(replaceWith('for ')).setName('loop trigram')
pass_stmt = Literal("<#>").setParseAction(replaceWith('pass')).setName('pass trigram')
cont_stmt = Literal("<$>").setParseAction(replaceWith('continue')).setName('continue trigram')
break_stmt = Literal("<%>").setParseAction(replaceWith('break')).setName('break trigram')
if_cond = Literal('<?>').setParseAction(replaceWith('if ')).setName('conditional trigram')
respects = Keyword("->").setParseAction(lambda s, l, t: ":").setName('contract assignment')
imposes = Keyword("<-").setParseAction(lambda s, l, t: "__new_contract").setName('contract imposition')
struct_main = Keyword("main:").setParseAction(replaceWith('if __name__ == "__main__":'))




from collections.abc import Hashable
import math
import pickle
import pytest
import uuid
from Aspidites._vendor.pyrsistent import (
    field, InvariantException, PClass, optional, CheckedPVector,
    pmap_field, pset_field, pvector_field)


class Point(PClass):
    x = field(type=int, mandatory=True, invariant=lambda x: (x >= 0, 'X negative'))
    y = field(type=int, serializer=lambda formatter, y: formatter(y))
    z = field(type=int, initial=0)


class Hierarchy(PClass):
    point = field(type=Point)


class TypedContainerObj(PClass):
    map = pmap_field(str, str)
    set = pset_field(str)
    vec = pvector_field(str)


class UniqueThing(PClass):
    id = field(type=uuid.UUID, factory=uuid.UUID)
    x = field(type=int)


def test_create_ignore_extra():
    p = Point.create({'x': 5, 'y': 10, 'z': 15, 'a': 0}, ignore_extra=True)
    assert p.x == 5
    assert p.y == 10
    assert p.z == 15
    assert isinstance(p, Point)


def test_create_ignore_extra_false():
    with pytest.raises(AttributeError):
        _ = Point.create({'x': 5, 'y': 10, 'z': 15, 'a': 0})


def test_create_ignore_extra_true():
    h = Hierarchy.create(
        {'point': {'x': 5, 'y': 10, 'z': 15, 'extra_field_0': 'extra_data_0'}, 'extra_field_1': 'extra_data_1'},
        ignore_extra=True)
    assert isinstance(h, Hierarchy)


def test_evolve_pclass_instance():
    p = Point(x=1, y=2)
    p2 = p.set(x=p.x + 2)

    # Original remains
    assert p.x == 1
    assert p.y == 2

    # Evolved object updated
    assert p2.x == 3
    assert p2.y == 2

    p3 = p2.set('x', 4)
    assert p3.x == 4
    assert p3.y == 2


def test_direct_assignment_not_possible():
    p = Point(x=1, y=2)

    with pytest.raises(AttributeError):
        p.x = 1

    with pytest.raises(AttributeError):
        setattr(p, 'x', 1)


def test_direct_delete_not_possible():
    p = Point(x=1, y=2)
    with pytest.raises(AttributeError):
        del p.x

    with pytest.raises(AttributeError):
        delattr(p, 'x')


def test_cannot_construct_with_undeclared_fields():
    with pytest.raises(AttributeError):
        Point(x=1, p=5)


def test_cannot_construct_with_wrong_type():
    with pytest.raises(TypeError):
        Point(x='a')


def test_cannot_construct_without_mandatory_fields():
    try:
        Point(y=1)
        assert False
    except InvariantException as e:
        assert "[Point.x]" in str(e)


def test_field_invariant_must_hold():
    try:
        Point(x=-1)
        assert False
    except InvariantException as e:
        assert "X negative" in str(e)


def test_initial_value_set_when_not_present_in_arguments():
    p = Point(x=1, y=2)

    assert p.z == 0


class Line(PClass):
    p1 = field(type=Point)
    p2 = field(type=Point)


def test_can_create_nested_structures_from_dict_and_serialize_back_to_dict():
    source = dict(p1=dict(x=1, y=2, z=3), p2=dict(x=10, y=20, z=30))
    l = Line.create(source)

    assert l.p1.x == 1
    assert l.p1.y == 2
    assert l.p1.z == 3
    assert l.p2.x == 10
    assert l.p2.y == 20
    assert l.p2.z == 30

    assert l.serialize(format=lambda val: val) == source


def test_can_serialize_with_custom_serializer():
    p = Point(x=1, y=1, z=1)

    assert p.serialize(format=lambda v: v + 17) == {'x': 1, 'y': 18, 'z': 1}


def test_implements_proper_equality_based_on_equality_of_fields():
    p1 = Point(x=1, y=2)
    p2 = Point(x=3)
    p3 = Point(x=1, y=2)

    assert p1 == p3
    assert not p1 != p3
    assert p1 != p2
    assert not p1 == p2


def test_is_hashable():
    p1 = Point(x=1, y=2)
    p2 = Point(x=3, y=2)

    d = {p1: 'A point', p2: 'Another point'}

    p1_like = Point(x=1, y=2)
    p2_like = Point(x=3, y=2)

    assert isinstance(p1, Hashable)
    assert d[p1_like] == 'A point'
    assert d[p2_like] == 'Another point'
    assert Point(x=10) not in d


def test_supports_nested_transformation():
    l1 = Line(p1=Point(x=2, y=1), p2=Point(x=20, y=10))

    l2 = l1.transform(['p1', 'x'], 3)

    assert l1.p1.x == 2

    assert l2.p1.x == 3
    assert l2.p1.y == 1
    assert l2.p2.x == 20
    assert l2.p2.y == 10


def test_repr():
    class ARecord(PClass):
        a = field()
        b = field()

    assert repr(ARecord(a=1, b=2)) in ('ARecord(a=1, b=2)', 'ARecord(b=2, a=1)')


def test_global_invariant_check():
    class UnitCirclePoint(PClass):
        __invariant__ = lambda cp: (0.99 < math.sqrt(cp.x * cp.x + cp.y * cp.y) < 1.01,
                                    "Point not on unit circle")
        x = field(type=float)
        y = field(type=float)

    UnitCirclePoint(x=1.0, y=0.0)

    with pytest.raises(InvariantException):
        UnitCirclePoint(x=1.0, y=1.0)


def test_supports_pickling():
    p1 = Point(x=2, y=1)
    p2 = pickle.loads(pickle.dumps(p1, -1))

    assert p1 == p2
    assert isinstance(p2, Point)


def test_supports_pickling_with_typed_container_fields():
    obj = TypedContainerObj(map={'foo': 'bar'}, set=['hello', 'there'], vec=['a', 'b'])
    obj2 = pickle.loads(pickle.dumps(obj))
    assert obj == obj2


def test_can_remove_optional_member():
    p1 = Point(x=1, y=2)
    p2 = p1.remove('y')

    assert p2 == Point(x=1)


def test_cannot_remove_mandatory_member():
    p1 = Point(x=1, y=2)

    with pytest.raises(InvariantException):
        p1.remove('x')


def test_cannot_remove_non_existing_member():
    p1 = Point(x=1)

    with pytest.raises(AttributeError):
        p1.remove('y')


def test_evolver_without_evolution_returns_original_instance():
    p1 = Point(x=1)
    e = p1.evolver()

    assert e.persistent() is p1


def test_evolver_with_evolution_to_same_element_returns_original_instance():
    p1 = Point(x=1)
    e = p1.evolver()
    e.set('x', p1.x)

    assert e.persistent() is p1


def test_evolver_supports_chained_set_and_remove():
    p1 = Point(x=1, y=2)

    assert p1.evolver().set('x', 3).remove('y').persistent() == Point(x=3)


def test_evolver_supports_dot_notation_for_setting_and_getting_elements():
    e = Point(x=1, y=2).evolver()

    e.x = 3
    assert e.x == 3
    assert e.persistent() == Point(x=3, y=2)


class Numbers(CheckedPVector):
    __type__ = int


class LinkedList(PClass):
    value = field(type='class_test.Numbers')
    next = field(type=optional('class_test.LinkedList'))


def test_string_as_type_specifier():
    l = LinkedList(value=[1, 2], next=LinkedList(value=[3, 4], next=None))

    assert isinstance(l.value, Numbers)
    assert list(l.value) == [1, 2]
    assert l.next.next is None


def test_multiple_invariants_on_field():
    # If the invariant returns a list of tests the results of running those tests will be
    # a tuple containing result data of all failing tests.

    class MultiInvariantField(PClass):
        one = field(type=int, invariant=lambda x: ((False, 'one_one'),
                                                   (False, 'one_two'),
                                                   (True, 'one_three')))
        two = field(invariant=lambda x: (False, 'two_one'))

    try:
        MultiInvariantField(one=1, two=2)
        assert False
    except InvariantException as e:
        assert set(e.invariant_errors) == set([('one_one', 'one_two'), 'two_one'])


def test_multiple_global_invariants():
    class MultiInvariantGlobal(PClass):
        __invariant__ = lambda self: ((False, 'x'), (False, 'y'))
        one = field()

    try:
        MultiInvariantGlobal(one=1)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == (('x', 'y'),)


def test_inherited_global_invariants():
    class Distant(object):
        def __invariant__(self):
            return [(self.distant, "distant")]

    class Nearby(Distant):
        def __invariant__(self):
            return [(self.nearby, "nearby")]

    class MultipleInvariantGlobal(Nearby, PClass):
        distant = field()
        nearby = field()

    try:
        MultipleInvariantGlobal(distant=False, nearby=False)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == (("nearby",), ("distant",),)


def test_diamond_inherited_global_invariants():
    counter = []

    class Base(object):
        def __invariant__(self):
            counter.append(None)
            return [(False, "base")]

    class Left(Base):
        pass

    class Right(Base):
        pass

    class SingleInvariantGlobal(Left, Right, PClass):
        pass

    try:
        SingleInvariantGlobal()
        assert False
    except InvariantException as e:
        assert e.invariant_errors == (("base",),)
        assert counter == [None]


def test_supports_weakref():
    import weakref
    weakref.ref(Point(x=1, y=2))


def test_supports_weakref_with_multi_level_inheritance():
    import weakref

    class PPoint(Point):
        a = field()

    weakref.ref(PPoint(x=1, y=2))


def test_supports_lazy_initial_value_for_field():
    class MyClass(PClass):
        a = field(int, initial=lambda: 2)

    assert MyClass() == MyClass(a=2)


def test_type_checks_lazy_initial_value_for_field():
    class MyClass(PClass):
        a = field(int, initial=lambda: "a")

    with pytest.raises(TypeError):
        MyClass()


def test_invariant_checks_lazy_initial_value_for_field():
    class MyClass(PClass):
        a = field(int, invariant=lambda x: (x < 5, "Too large"), initial=lambda: 10)

    with pytest.raises(InvariantException):
        MyClass()


def test_invariant_checks_static_initial_value():
    class MyClass(PClass):
        a = field(int, invariant=lambda x: (x < 5, "Too large"), initial=10)

    with pytest.raises(InvariantException):
        MyClass()


def test_lazy_invariant_message():
    class MyClass(PClass):
        a = field(int, invariant=lambda x: (x < 5, lambda: "{x} is too large".format(x=x)))

    try:
        MyClass(a=5)
        assert False
    except InvariantException as e:
        assert '5 is too large' in e.invariant_errors


def test_enum_key_type():
    import enum
    class Foo(enum.Enum):
        Bar = 1
        Baz = 2

    # This currently fails because the enum is iterable
    class MyClass1(PClass):
        f = pmap_field(key_type=Foo, value_type=int)

    MyClass1()

    # This is OK since it's wrapped in a tuple
    class MyClass2(PClass):
        f = pmap_field(key_type=(Foo,), value_type=int)

    MyClass2()


def test_pickle_with_one_way_factory():
    thing = UniqueThing(id='25544626-86da-4bce-b6b6-9186c0804d64')
    assert pickle.loads(pickle.dumps(thing)) == thing


def test_evolver_with_one_way_factory():
    thing = UniqueThing(id='cc65249a-56fe-4995-8719-ea02e124b234')
    ev = thing.evolver()
    ev.x = 5  # necessary to prevent persistent() returning the original
    assert ev.persistent() == UniqueThing(id=str(thing.id), x=5)


def test_set_doesnt_trigger_other_factories():
    thing = UniqueThing(id='b413b280-de76-4e28-a8e3-5470ca83ea2c')
    thing.set(x=5)


def test_set_does_trigger_factories():
    class SquaredPoint(PClass):
        x = field(factory=lambda x: x ** 2)
        y = field()

    sp = SquaredPoint(x=3, y=10)
    assert (sp.x, sp.y) == (9, 10)

    sp2 = sp.set(x=4)
    assert (sp2.x, sp2.y) == (16, 10)


def test_value_can_be_overridden_in_subclass_new():
    class X(PClass):
        y = pvector_field(int)

        def __new__(cls, **kwargs):
            items = kwargs.get('y', None)
            if items is None:
                kwargs['y'] = ()
            return super(X, cls).__new__(cls, **kwargs)

    a = X(y=[])
    b = a.set(y=None)
    assert a == b



# pgn.py rel. 1.1 17-sep-2004
#
# Demonstration of the parsing module, implementing a pgn parser.
#
# The aim of this parser is not to support database application,
# but to create automagically a pgn annotated reading the log console file
# of a lecture of ICC (Internet Chess Club), saved by Blitzin.
# Of course you can modify the Abstract Syntax Tree to your purpose.
#
# Copyright 2004, by Alberto Santini http://www.albertosantini.it/chess/
#
from pyparsing import alphanums, nums, quotedString
from pyparsing import Combine, Forward, Group, Literal, oneOf, OneOrMore, Optional, Suppress, ZeroOrMore, Word
from pyparsing import ParseException

#
# define pgn grammar
#

tag = Suppress("[") + Word(alphanums) + Combine(quotedString) + Suppress("]")
comment = Suppress("{") + Word(alphanums + " ") + Suppress("}")

dot = Literal(".")
piece = oneOf("K Q B N R")
file_coord = oneOf("a b c d e f g h")
rank_coord = oneOf("1 2 3 4 5 6 7 8")
capture = oneOf("x :")
promote = Literal("=")
castle_queenside = oneOf("O-O-O 0-0-0 o-o-o")
castle_kingside = oneOf("O-O 0-0 o-o")

move_number = Optional(comment) + Word(nums) + dot
m1 = file_coord + rank_coord # pawn move e.g. d4
m2 = file_coord + capture + file_coord + rank_coord # pawn capture move e.g. dxe5
m3 = file_coord + "8" + promote + piece # pawn promotion e.g. e8=Q
m4 = piece + file_coord + rank_coord # piece move e.g. Be6
m5 = piece + file_coord + file_coord + rank_coord # piece move e.g. Nbd2
m6 = piece + rank_coord + file_coord + rank_coord # piece move e.g. R4a7
m7 = piece + capture + file_coord + rank_coord # piece capture move e.g. Bxh7
m8 = castle_queenside | castle_kingside # castling e.g. o-o

check = oneOf("+ ++")
mate = Literal("#")
annotation = Word("!?", max=2)
nag = " $" + Word(nums)
decoration = check | mate | annotation | nag

variant = Forward()
half_move = Combine((m3 | m1 | m2 | m4 | m5 | m6 | m7 | m8) + Optional(decoration)) \
  + Optional(comment) +Optional(variant)
move = Suppress(move_number) + half_move + Optional(half_move)
variant << "(" + OneOrMore(move) + ")"
# grouping the plies (half-moves) for each move: useful to group annotations, variants...
# suggested by Paul McGuire :)
move = Group(Suppress(move_number) + half_move + Optional(half_move))
variant << Group("(" + OneOrMore(move) + ")")
game_terminator = oneOf("1-0 0-1 1/2-1/2 *")

pgnGrammar = Suppress(ZeroOrMore(tag))  + ZeroOrMore(move) + Optional(Suppress(game_terminator))

def parsePGN( pgn, bnf=pgnGrammar, fn=None ):
  try:
    return bnf.parseString( pgn )
  except ParseException as err:
    print(err.line)
    print(" "*(err.column-1) + "^")
    print(err)

if __name__ == "__main__":
  # input string
  pgn = """
[Event "ICC 5 0 u"]
[Site "Internet Chess Club"]
[Date "2004.01.25"]
[Round "-"]
[White "guest920"]
[Black "IceBox"]
[Result "0-1"]
[ICCResult "White checkmated"]
[BlackElo "1498"]
[Opening "French defense"]
[ECO "C00"]
[NIC "FR.01"]
[Time "04:44:56"]
[TimeControl "300+0"]

1. e4 e6 2. Nf3 d5 $2 3. exd5 (3. e5 g6 4. h4) exd5 4. Qe2+ Qe7 5. Qxe7+ Bxe7 6. d3 Nf6 7. Be3
Bg4 8. Nbd2 c5 9. h3 Be6 10. O-O-O Nc6 11. g4 Bd6 12. g5 Nd7 13. Rg1 d4 14.
g6 fxg6 15. Bg5 Rf8 16. a3 Bd5 17. Re1+ Nde5 18. Nxe5 Nxe5 19. Bf4 Rf5 20.
Bxe5 Rxe5 21. Rg5 Rxe1# {Black wins} 0-1
"""
  # parse input string
  tokens = parsePGN(pgn, pgnGrammar)
  print(tokens.dump())



import unittest

from Aspidites._vendor.contracts import (check, ContractNotRespected, Contract, parse,
                                         check_multiple, ContractSyntaxError, fail)


class TestIdioms(unittest.TestCase):

    def test_check_1(self):
        res = check('tuple(int,int)', (2, 2))

        assert isinstance(res, dict)

    def test_check_1a(self):
        self.assertRaises(ValueError, check, 1, 2)

    def test_parse_1(self):
        contract = parse('>0')
        assert isinstance(contract, Contract)
        contract.check(2)
        self.assertRaises(ContractNotRespected, contract.check, 'ciao')

    def test_parse_2(self):
        self.assertRaises(ContractSyntaxError, parse, '>>')

    def test_check_2(self):
        self.assertRaises(ContractNotRespected,
                          check, 'tuple(int,int)', (None, 2))

    def test_check_3(self):
        self.assertRaises(ContractSyntaxError,
                          check, 'tuple(>>int,int)', (None, 2))

    def test_check_4(self):
        score = (2, None)
        msg = 'Player score must be a tuple of 2 int.'
        try:
            check('tuple(int,int)', score, msg)
        except ContractNotRespected as e:
            s = str(e)
            assert msg in s
        else:
            assert False

    def test_repr_1(self):
        contract = parse(' list[N](int), N > 0')

        ("%s" % contract)   # => 'list[N](int),N>0'
        ("%r" % contract)   # => And([List(BindVariable('N',int),...

    def test_binding(self):
        context = check('list[N](str), N>0', ['a', 'b', 'c'])

        self.assertTrue('N' in context)
        self.assertTrue(context['N'] == 3)

    def test_check_multiple_1(self):

        data = [[1, 2, 3],
                [4, 5, 6]]
        row_labels = ['first season', 'second season']
        col_labels = ['Team1', 'Team2', 'Team3']

        spec = [('list[C](str),C>0', col_labels),
                ('list[R](str),R>0', row_labels),
                ('list[R](list[C])', data)]
        check_multiple(spec)

        # now with description 
        check_multiple(spec,
                        'I expect col_labels, row_labels, data to '
                        'have coherent dimensions.')

        data = [[1, 2, 3], [1, 2]]
        spec = [('list[C](str),C>0', col_labels),
                ('list[R](str),R>0', row_labels),
                ('list[R](list[C])', data)]

        self.assertRaises(ContractNotRespected, check_multiple, spec)
        self.assertRaises(ContractNotRespected, check_multiple, spec,
                          'my message')

#     def test_symbols(self):
#         from contracts import contract_expression  # @UnusedImport


    def test_equality_contract(self):
        c1 = parse('list[C](str),C>0')
        c2 = parse('list[C](str),C>0')
        c3 = parse('list[R](str),R>0')
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, c3)

    def test_equality_rvalue(self):
        c1 = parse('N+1')
        c2 = parse('N+2')
        c3 = parse('P+1')
        self.assertEqual(c1, c1)
        self.assertEqual(c2, c2)
        self.assertEqual(c3, c3)
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c3)
        self.assertNotEqual(c2, c3)

    def test_check_context(self):
        check('N', 1, N=1)
        fail('N', 1, N=2)

        self.assertRaises(ContractNotRespected, check, 'N', 1, N=2)

        self.assertRaises(ValueError, fail, 'N', 1, N=1)

    def test_check_context2(self):
        """ Variable names must have only one letter. """
        self.assertRaises(ValueError, check, 'N', 1, NN=2)
        self.assertRaises(ValueError, check, 'N', 1, nn=2)




from . import good, fail
import hypothesis
import hypothesis.strategies as st
import pytest as pt
from Aspidites._vendor.pyrsistent import pvector, pmap, pset


fail('Container', 1)
good('Hashable', 1)
fail('Iterable', 1)
fail('Sized', lambda: None)
good('Sequence', '')
good('Sized', '')
good('Callable', lambda: None)


@hypothesis.given(x=st.lists(st.text() | st.binary()))
def random_sequence(x):
    good('Container', [*x])
    good('Hashable', pvector([*x]))
    # counterexample?
    good('Iterable', [*x])
    good('Iterable', pvector([*x]))
    good('Iterator', [*x].__iter__())
    fail('Iterator', [*x])
    good('Sized', [*x])
    good('Sized', pvector([*x]))
    fail('Callable', [*x])
    good('Sequence', [*x])
    good('Sequence', pvector([*x]))
    good('Sequence', set(x))
    good('Set', {*x})
    good('Set', pset([*x]))
    fail('Set', [*x])
    good('MutableSequence', [*x])
    good('MutableSequence', pvector([*x]).evolver())
    fail('MutableSequence', (*x, ))
    good('MutableSet', {*x})
    good('MutableSet', pset([*x]).evolver())
    fail('MutableSet', pset([*x]))
    fail('MutableSet', frozenset([*x]))


random_sequence()


@hypothesis.given(x=st.dictionaries(st.text(), st.binary()) | st.fixed_dictionaries({}))
def random_mappings(x):
    good('Mapping', {**x})
    good('Mapping', pmap({**x}))
    fail('Mapping', [x])
    fail('Mapping', pvector([x]))
    good('MutableMapping', {**x})
    good('MutableMapping', pmap({**x}).evolver())
    good('Iterable', pmap({**x}))
    good('Iterable', {**x})
    fail('Sequence', {**x})
    fail('Sequence', pmap({**x}))
    good('Sized', pmap({**x}))
    good('Sized', {**x})
    # good('MappingView', {}.keys())
    good('Sized', {**x}.keys())
    good('Sized', pmap({**x}).keys())
    # good('ItemsView', {}.items())
    good('Sized', {**x}.items())
    good('Sized', pmap({**x}).items())
    # good('ValuesView', {}.values())
    good('Sized', {**x}.values())
    good('Sized', pmap({**x}).values())


random_mappings()



# -*- coding: utf-8 -*-

from Aspidites._vendor.pyrsistent._pmap import pmap, m, PMap

from Aspidites._vendor.pyrsistent._pvector import pvector, v, PVector

from Aspidites._vendor.pyrsistent._pset import pset, s, PSet

from Aspidites._vendor.pyrsistent._pbag import pbag, b, PBag

from Aspidites._vendor.pyrsistent._plist import plist, l, PList

from Aspidites._vendor.pyrsistent._pdeque import pdeque, dq, PDeque

from Aspidites._vendor.pyrsistent._checked_types import (
    CheckedPMap, CheckedPVector, CheckedPSet, InvariantException, CheckedKeyTypeError,
    CheckedValueTypeError, CheckedType, optional)

from Aspidites._vendor.pyrsistent._field_common import (
    field, PTypeError, pset_field, pmap_field, pvector_field)

from Aspidites._vendor.pyrsistent._precord import PRecord

from Aspidites._vendor.pyrsistent._pclass import PClass, PClassMeta

from Aspidites._vendor.pyrsistent._immutable import immutable

from Aspidites._vendor.pyrsistent._helpers import freeze, thaw, mutant

from Aspidites._vendor.pyrsistent._transformations import inc, discard, rex, ny

from Aspidites._vendor.pyrsistent._toolz import get_in

__version__ = '0.18.0'

__all__ = ('pmap', 'm', 'PMap',
           'pvector', 'v', 'PVector',
           'pset', 's', 'PSet',
           'pbag', 'b', 'PBag',
           'plist', 'l', 'PList',
           'pdeque', 'dq', 'PDeque',
           'CheckedPMap', 'CheckedPVector', 'CheckedPSet', 'InvariantException', 'CheckedKeyTypeError', 'CheckedValueTypeError', 'CheckedType', 'optional',
           'PRecord', 'field', 'pset_field', 'pmap_field', 'pvector_field',
           'PClass', 'PClassMeta',
           'immutable',
           'freeze', 'thaw', 'mutant',
           'get_in',
           'inc', 'discard', 'rex', 'ny')



import datetime
import pickle
import pytest
from Aspidites._vendor.pyrsistent import CheckedPVector, InvariantException, optional, CheckedValueTypeError, PVector


class Naturals(CheckedPVector):
    __type__ = int
    __invariant__ = lambda value: (value >= 0, 'Negative value')

def test_instantiate():
    x = Naturals([1, 2, 3])

    assert list(x) == [1, 2, 3]
    assert isinstance(x, Naturals)
    assert isinstance(x, PVector)

def test_append():
    x = Naturals()
    x2 = x.append(1)

    assert list(x2) == [1]
    assert isinstance(x2, Naturals)

def test_extend():
    x = Naturals()
    x2 = x.extend([1])

    assert list(x2) == [1]
    assert isinstance(x2, Naturals)

def test_set():
    x = Naturals([1, 2])
    x2 = x.set(1, 3)

    assert list(x2) == [1, 3]
    assert isinstance(x2, Naturals)


def test_invalid_type():
    try:
        Naturals([1, 2.0])
        assert False
    except CheckedValueTypeError as e:
        assert e.expected_types == (int,)
        assert e.actual_type is float
        assert e.actual_value == 2.0
        assert e.source_class is Naturals

    x = Naturals([1, 2])
    with pytest.raises(TypeError):
        x.append(3.0)

    with pytest.raises(TypeError):
        x.extend([3, 4.0])

    with pytest.raises(TypeError):
        x.set(1, 2.0)

    with pytest.raises(TypeError):
        x.evolver()[1] = 2.0

def test_breaking_invariant():
    try:
        Naturals([1, -1])
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ('Negative value',)

    x = Naturals([1, 2])
    try:
        x.append(-1)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ('Negative value',)

    try:
        x.extend([-1])
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ('Negative value',)

    try:
        x.set(1, -1)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ('Negative value',)

def test_create_base_case():
    x =  Naturals.create([1, 2, 3])

    assert isinstance(x, Naturals)
    assert x == Naturals([1, 2, 3])

def test_create_with_instance_of_checked_pvector_returns_the_argument():
    x =  Naturals([1, 2, 3])

    assert Naturals.create(x) is x

class OptionalNaturals(CheckedPVector):
    __type__ = optional(int)
    __invariant__ = lambda value: (value is None or value >= 0, 'Negative value')

def test_multiple_allowed_types():
    assert list(OptionalNaturals([1, None, 3])) == [1, None, 3]

class NaturalsVector(CheckedPVector):
    __type__ = optional(Naturals)

def test_create_of_nested_structure():
    assert NaturalsVector([Naturals([1, 2]), Naturals([3, 4]), None]) ==\
           NaturalsVector.create([[1, 2], [3, 4], None])

def test_serialize_default_case():
    v = CheckedPVector([1, 2, 3])
    assert v.serialize() == [1, 2, 3]

class Dates(CheckedPVector):
    __type__ = datetime.date

    @staticmethod
    def __serializer__(format, d):
        return d.strftime(format)

def test_serialize_custom_serializer():
    d = datetime.date
    v = Dates([d(2015, 2, 2), d(2015, 2, 3)])
    assert v.serialize(format='%Y-%m-%d') == ['2015-02-02', '2015-02-03']

def test_type_information_is_inherited():
    class MultiDates(Dates):
        __type__ = int

    MultiDates([datetime.date(2015, 2, 4), 5])

    with pytest.raises(TypeError):
        MultiDates([5.0])

def test_invariants_are_inherited():
    class LimitNaturals(Naturals):
        __invariant__ = lambda value: (value < 10, 'Too big')

    try:
        LimitNaturals([10, -1])
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ('Too big', 'Negative value')

def test_invariant_must_be_callable():
    with pytest.raises(TypeError):
        class InvalidInvariant(CheckedPVector):
            __invariant__ = 1

def test_type_spec_must_be_type():
    with pytest.raises(TypeError):
        class InvalidType(CheckedPVector):
            __type__ = 1

def test_repr():
    x = Naturals([1, 2])

    assert str(x) == 'Naturals([1, 2])'

def test_evolver_returns_same_instance_when_no_updates():
    x = Naturals([1, 2])
    assert x.evolver().persistent() is x

def test_pickling():
    x = Naturals([1, 2])
    y = pickle.loads(pickle.dumps(x, -1))

    assert x == y
    assert isinstance(y, Naturals)

def test_multiple_optional_types():
    class Numbers(CheckedPVector):
        __type__ = optional(int, float)

    numbers = Numbers([1, 2.5, None])
    assert numbers.serialize() == [1, 2.5, None]

    with pytest.raises(TypeError):
        numbers.append('foo')


class NaturalsVectorStr(CheckedPVector):
    __type__ = 'checked_vector_test.Naturals'


def test_check_with_string_specification():
    naturals_list = [Naturals([1, 2]), Naturals([3, 4])]
    nv = NaturalsVectorStr(naturals_list)
    assert nv == naturals_list


def test_create_with_string_specification():
    naturals_list = [[1, 2], [3, 4]]
    nv = NaturalsVectorStr.create(naturals_list)
    assert nv == naturals_list


def test_supports_weakref():
    import weakref
    weakref.ref(Naturals([]))


def test_create_with_generator_iterator():
    # See issue #97
    class Numbers(CheckedPVector):
        __type__ = int

    n = Numbers(i for i in [1, 2, 3])
    assert n == Numbers([1, 2, 3])


import inspect
from inspect import CO_VARARGS  # pylint: disable=no-name-in-module
from itertools import chain, repeat
from types import CodeType
from typing import List, Optional, Type, get_type_hints, Dict, Union, Mapping, Iterable


def get_arg_types(obj) -> List[Optional[Type]]:
    code: CodeType = obj.__code__
    args = []
    argcount = code.co_argcount
    if code.co_flags & CO_VARARGS:
        argcount += 1
    type_hints = get_type_hints(obj)
    for ix, name in zip(range(0, argcount), code.co_varnames):
        args.append(type_hints.get(name, None))
    return args


def is_keyword_parameter(p: inspect.Parameter) -> bool:
    if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
        return True
    if p.kind == inspect.Parameter.KEYWORD_ONLY:
        return True
    return False


def get_kwarg_types(obj) -> Dict[str, type]:
    sig: inspect.Signature = inspect.signature(obj)
    type_hints = get_type_hints(obj)
    result = {}
    for name, parameter in sig.parameters.items():
        if not is_keyword_parameter(parameter):
            continue
        result[parameter.name] = type_hints.get(parameter.name, None)
    return result


def get_return_type(obj) -> Optional[Type]:
    type_hints = get_type_hints(obj)
    return type_hints.get('return', None)


def invoke(func, args: Union[Mapping, Iterable]):
    code: CodeType = func.__code__
    argcount = code.co_argcount
    actual_args = []
    if isinstance(args, Mapping):
        for _, name in zip(range(0, argcount), code.co_varnames):
            arg = args[name] if name in args else None
            actual_args.append(arg)
    else:
        for _, arg in zip(range(0, argcount), chain(args, repeat(None))):
            actual_args.append(arg)

    return func(*actual_args)



#
# cpp_enum_parser.py
#
# Posted by Mark Tolonen on comp.lang.python in August, 2009,
# Used with permission.
#
# Parser that scans through C or C++ code for enum definitions, and
# generates corresponding Python constant definitions.
#
#

from ....._vendor.pyparsing import *
# sample string with enums and other stuff
sample = '''
    stuff before
    enum hello {
        Zero,
        One,
        Two,
        Three,
        Five=5,
        Six,
        Ten=10
        };
    in the middle
    enum blah
        {
        alpha,
        beta,
        gamma = 10 ,
        zeta = 50
        };
    at the end
    '''

# syntax we don't want to see in the final parse tree
LBRACE,RBRACE,EQ,COMMA = map(Suppress,"{}=,")
_enum = Suppress('enum')
identifier = Word(alphas,alphanums+'_')
integer = Word(nums)
enumValue = Group(identifier('name') + Optional(EQ + integer('value')))
enumList = Group(enumValue + ZeroOrMore(COMMA + enumValue))
enum = _enum + identifier('enum') + LBRACE + enumList('names') + RBRACE

# find instances of enums ignoring other syntax
for item,start,stop in enum.scanString(sample):
    id = 0
    for entry in item.names:
        if entry.value != '':
            id = int(entry.value)
        print('%s_%s = %d' % (item.enum.upper(),entry.name.upper(),id))
        id += 1



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
import numpy as np
from Aspidites._vendor.contracts import new_contract
from Aspidites._vendor.contracts.interface import describe_value, describe_type
__all__ = ['finite']


@new_contract
def finite(x):
    return np.isfinite(x).all()


new_contract('np_scalar_uint', 'np_uint8|np_uint16|np_uint32|np_uint64')
new_contract('np_scalar_int', 'np_int8|np_int16|np_int32|np_int64')
new_contract('np_scalar_float', 'np_float32|np_float64')
new_contract('np_scalar_type', 'np_scalar_int|np_scalar_uint|np_scalar_float')


@new_contract
def np_zeroshape_array(x):
    scalars: tuple = (
    np.ndarray,
    np.int,  # Platform integer (normally either int32 or int64)
    np.int8,  # Byte (-128 to 127)
    np.int16,  # Integer (-32768 to 32767)
    np.int32,  # Integer (-2147483648 to 2147483647)
    np.int64,  # Integer (-9223372036854775808 to 9223372036854775807)
    np.uint8,  # Unsigned integer (0 to 255)
    np.uint16,  # Unsigned integer (0 to 65535)
    np.uint32,  # Unsigned integer (0 to 4294967295)
    np.uint64,  # Unsigned integer (0 to 18446744073709551615)
    np.float,  # Shorthand for float64.
    np.float16,  #  Half precision float: sign bit, 5 bits exponent, 10 bits mantissa
    np.float32,  #  Single precision float: sign bit, 8 bits exponent, 23 bits mantissa
    np.float64,  #  Double precision float: sign bit, 11 bits exponent, 52 bits mantissa
    np.complex,  #  Shorthand for complex128.
    np.complex64,  #    Complex number, represented by two 32-bit floats (real and imaginary components)
    np.complex128
    )

    if not isinstance(x, np.ndarray):
        msg = 'Not an array: %s %s ' % (type(x), describe_type(x))
        raise ValueError(msg)

    if not x.shape == ():
        msg = 'Not a scalar: %s' % describe_value(x)
        raise ValueError(msg)

    return True


new_contract('np_scalar', 'np_zeroshape_array|np_scalar_type')




#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected, describe_type
from ..pyparsing_utils import myOperatorPrecedence
from ..syntax import (add_contract, W, contract_expression, O, S, rvalue,
    simple_contract, ZeroOrMore, Literal, MatchFirst, opAssoc, FollowedBy, NotAny,
    Keyword, add_keyword, Word)
from .array_ops import (ArrayOR, ArrayAnd, DType, ArrayConstraint,
    ArrayORCustomString)
from .compositions import And, OR
from .suggester import create_suggester
from numpy import ndarray, dtype
import numpy
from Aspidites._vendor.pyparsing import infixNotation as operatorPrecedence, Or


class Array(Contract):

    def __init__(self, shape_contract=None,
                        elements_contract=None, where=None):
        Contract.__init__(self, where)
        self.shape_contract = shape_contract
        self.elements_contract = elements_contract

    def check_contract(self, context, value, silent):
        if not isinstance(value, ndarray):
            error = 'Expected an array, got a %s.' % describe_type(value)
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

        if self.shape_contract is not None:
            self.shape_contract._check_contract(context, value.shape, silent)

        if self.elements_contract is not None:
            self.elements_contract._check_contract(context, value, silent)

    def __str__(self):
        s = 'array'
        if self.shape_contract is not None:
            s += '[%s]' % self.shape_contract
        if self.elements_contract is not None:
            s += '(%s)' % self.elements_contract
        return s

    def __repr__(self):
        s = 'Array(%r,%r)' % (self.shape_contract, self.elements_contract)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        shape_contract = tokens.get('shape_contract', [None])[0]
        elements_contract = tokens.get('elements_contract', [None])[0]

        assert shape_contract is None or isinstance(shape_contract,
                                                    ShapeContract)
        assert elements_contract is None or isinstance(elements_contract,
                                                       Contract)
        return Array(shape_contract, elements_contract, where=where)


class ShapeContract(Contract):
    def __init__(self, dimensions, ellipsis=False, where=None):
        assert isinstance(dimensions, list)
        assert isinstance(ellipsis, bool)
        Contract.__init__(self, where)
        self.dimensions = dimensions
        self.ellipsis = ellipsis

    def check_contract(self, context, value, silent):
        assert isinstance(value, tuple)  # Guaranteed by construction

        expected = len(self.dimensions)
        ndim = len(value)

        if ndim < expected:
            error = 'Expected %d dimensions, got %d.' % (expected, ndim)
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

        if ndim > expected and not self.ellipsis:
            error = 'Expected %d dimensions, got %d.' % (expected, ndim)
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

        for i in range(expected):
            self.dimensions[i]._check_contract(context, value[i], silent)

    def __str__(self):
        be_careful = self.ellipsis or len(self.dimensions) > 1

        def rep(x):
            if be_careful and isinstance(x, (And, OR)):
                return "(%s)" % x
            else:
                return "%s" % x

        s = 'x'.join(rep(x) for x in self.dimensions)
        if self.ellipsis:
            s += 'x...'
        return s

    def __repr__(self):
        if self.ellipsis:
            s = 'ShapeContract(%r,%r)' % (self.dimensions, self.ellipsis)
        else:
            s = 'ShapeContract(%r)' % self.dimensions
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        # print "in tokens: " % tokens
        # workaround for some bugs
        ellipsis = False
        dimensions = []
        for t in tokens:
            if t == '...':
                ellipsis = True
            else:
                assert isinstance(t, Contract), 'Wrong token %r' % t
                dimensions.append(t)
        return ShapeContract(dimensions, ellipsis, where=where)


class Shape(Contract):
    def __init__(self, length, contract, where=None):
        Contract.__init__(self, where)
        self.contract = contract
        self.length = length

    def check_contract(self, context, value, silent):
        if not isinstance(value, ndarray):
            error = 'Expected an array, got %r.' % value.__class__.__name__
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

        if isinstance(value, ndarray):
            value = value.shape

        if self.length is not None:
            self.length._check_contract(context, len(value), silent)

        if self.contract is not None:
            self.contract._check_contract(context, value, silent)

    def __str__(self):
        s = 'shape'
        if self.length is not None:
            s += '[%s]' % self.length
        if self.contract is not None:
            s += '(%s)' % self.contract
        return s

    def __repr__(self):
        s = 'Shape(%r,%r)' % (self.length, self.contract)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        assert 0 <= len(tokens) <= 2
        length = tokens.get('length', [None])[0]
        contract = tokens.get('other', [None])[0]
        return Shape(length, contract, where)


array_constraints = []
for glyph in ArrayConstraint.constraints:
    if glyph == '!=':
        # special case: ! must be followed by =
        glyph_expression = Literal('!') - Literal('=')
        glyph_expression.setName('!=')
    else:
        glyph_expression = Literal(glyph)

    expr = glyph_expression('glyph') - rvalue('rvalue')
    expr.setParseAction(ArrayConstraint.parse_action)
    array_constraints.append(expr)

np_uint_dtypes = "u1 uint8 uint16 uint32 uint64".split()
np_int_dtypes = "i1 int8 int16 int32 int64".split()
np_float_dtypes = "float32 float64".split()
np_other_dtypes = ['bool']
atomic = np_uint_dtypes + np_int_dtypes + np_float_dtypes + np_other_dtypes

# in numpy, int = int64, float = float64
# for us, int = int64|int32| ...

dtype_checks = []
for x in atomic:
    d = numpy.dtype(x)
    expr = Keyword(x).setParseAction(DType.parse_action(d))
    dtype_checks.append(expr)

def np_composite(custom_string, alternatives):
    alts = [DType(numpy.dtype(a), a) for a in alternatives]
    return ArrayORCustomString(custom_string=custom_string, clauses=alts)

def np_uint(s, loc, tokens):
    return np_composite('uint', np_uint_dtypes)

def np_int(s, loc, tokens):
    return np_composite('int', np_int_dtypes)

def np_float(s, loc, tokens):
    return np_composite('float', np_float_dtypes)

dtype_checks.append(Keyword('int').setParseAction(np_int))
dtype_checks.append(Keyword('uint').setParseAction(np_uint))
dtype_checks.append(Keyword('float').setParseAction(np_float))

composite = ['int', 'uint', 'float']

ndarray_simple_contract = Or(dtype_checks + array_constraints)
ndarray_simple_contract.setName('numpy element contract')

suggester = create_suggester(get_options=lambda: atomic + composite)
baseExpr = ndarray_simple_contract | suggester
baseExpr.setName('numpy contract (with recovery)')

op = myOperatorPrecedence
# op = operatorPrecedence
ndarray_composite_contract = op(baseExpr, [
    (',', 2, opAssoc.LEFT, ArrayAnd.parse_action),
    ('|', 2, opAssoc.LEFT, ArrayOR.parse_action),
])


def my_delim_list2(what, delim):
    return (what + ZeroOrMore(S(delim) + FollowedBy(NotAny(ellipsis)) - what))


ellipsis = Literal('...')

shape_suggester = create_suggester(get_options=lambda: ['...'],
                                   pattern=Word('.'))

inside_inside1 = simple_contract | shape_suggester
inside_inside2 = contract_expression | shape_suggester
inside = (S('(') - inside_inside2 - S(')')) | inside_inside1  # XXX: ^ and use or_contract?
shape_contract = my_delim_list2(inside, S('x')) + O(S('x') + ellipsis)
shape_contract.setParseAction(ShapeContract.parse_action)
shape_contract.setName('array shape contract')

name = Keyword('array') | Keyword('ndarray')
optional_shape = (S('[') - shape_contract - S(']'))('shape_contract')
optional_elements = (S('(') - ndarray_composite_contract - S(')'))('elements_contract')
array_contract = name + O(optional_shape) + O(optional_elements)
array_contract.setParseAction(Array.parse_action)
array_contract.setName('array() contract')
add_contract(array_contract)
add_keyword('array')
add_keyword('ndarray')

optional_length = (S('[') - contract_expression - S(']'))('length')
optional_other = (S('(') - contract_expression - S(')'))('other')
shape = Keyword('shape') + O(optional_length) + O(optional_other)
shape.setName('shape() contract')
shape.setParseAction(Shape.parse_action)
add_contract(shape)
add_keyword('shape')







from ._vendor.pyrsistent import PClass


def safer_type(*args):
    name: str
    bases: tuple
    attributes: dict
    if len(args) < 2:
        raise RuntimeError("Type introspection is not supported.")
    elif len(args) == 2:
        name, attributes = args
        bases = (PClass,)
    else:
        name, bases, attributes = args
        bases = (PClass,) + bases

    return type(name, bases, dict(attributes))



"""This is the standard library bootstrapped with the compiler on a per-version basis.
This should allow for some semblance of backwards compatibility.
"""
# Aspidites - Woma Programming Language Standard Library
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .gcutils import get_all
from .fileutils import mkdir_p, atomic_save, iglob, iter_find_files, copytree
from .pickleutils import pickle_loads
from .guiutils import *
from .mathutils import clamp, ceil, floor
from contextlib import contextmanager
from functools import lru_cache

__version__ = '1.14.1'


def woma_version():
    return __version__



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec


SIMPLE_SUBSCRIPTS = """
def simple_subscript(a):
    return a['b']
"""


def test_read_simple_subscript(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(SIMPLE_SUBSCRIPTS, glb)

    assert (value, 'b') == glb['simple_subscript'](value)


VAR_SUBSCRIPT = """
def simple_subscript(a, b):
    return a[b]
"""


def test_read_subscript_with_variable(mocker):
    value = [1]
    idx = 0
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(VAR_SUBSCRIPT, glb)

    assert (value, 0) == glb['simple_subscript'](value, idx)


TUPLE_SUBSCRIPTS = """
def tuple_subscript(a):
    return a[1, 2]
"""


def test_tuple_subscript(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(TUPLE_SUBSCRIPTS, glb)

    assert (value, (1, 2)) == glb['tuple_subscript'](value)


SLICE_SUBSCRIPT_NO_UPPER_BOUND = """
def slice_subscript_no_upper_bound(a):
    return a[1:]
"""


def test_read_slice_subscript_no_upper_bound(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_NO_UPPER_BOUND, glb)

    assert (value, slice(1, None, None)) == glb['slice_subscript_no_upper_bound'](value)  # NOQA: E501


SLICE_SUBSCRIPT_NO_LOWER_BOUND = """
def slice_subscript_no_lower_bound(a):
    return a[:1]
"""


def test_read_slice_subscript_no_lower_bound(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_NO_LOWER_BOUND, glb)

    assert (value, slice(None, 1, None)) == glb['slice_subscript_no_lower_bound'](value)  # NOQA: E501


SLICE_SUBSCRIPT_NO_STEP = """
def slice_subscript_no_step(a):
    return a[1:2]
"""


def test_read_slice_subscript_no_step(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_NO_STEP, glb)

    assert (value, slice(1, 2, None)) == glb['slice_subscript_no_step'](value)


SLICE_SUBSCRIPT_WITH_STEP = """
def slice_subscript_with_step(a):
    return a[1:2:3]
"""


def test_read_slice_subscript_with_step(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_WITH_STEP, glb)

    assert (value, slice(1, 2, 3)) == glb['slice_subscript_with_step'](value)


EXTENDED_SLICE_SUBSCRIPT = """

def extended_slice_subscript(a):
    return a[0, :1, 1:, 1:2, 1:2:3]
"""


def test_read_extended_slice_subscript(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {'_getitem_': _getitem_}
    restricted_exec(EXTENDED_SLICE_SUBSCRIPT, glb)
    ret = glb['extended_slice_subscript'](value)
    ref = (
        value,
        (
            0,
            slice(None, 1, None),
            slice(1, None, None),
            slice(1, 2, None),
            slice(1, 2, 3)
        )
    )

    assert ref == ret


WRITE_SUBSCRIPTS = """
def assign_subscript(a):
    a['b'] = 1
"""


def test_write_subscripts(
        mocker):
    value = {'b': None}
    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob
    glb = {'_write_': _write_}
    restricted_exec(WRITE_SUBSCRIPTS, glb)

    glb['assign_subscript'](value)
    assert value['b'] == 1


DEL_SUBSCRIPT = """
def del_subscript(a):
    del a['b']
"""


def test_del_subscripts(
        mocker):
    value = {'b': None}
    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob
    glb = {'_write_': _write_}
    restricted_exec(DEL_SUBSCRIPT, glb)
    glb['del_subscript'](value)

    assert value == {}



from enum import Enum

from Aspidites._vendor.pyrsistent import field, pvector_field


class ExampleEnum(Enum):
    x = 1
    y = 2


def test_enum():
    f = field(type=ExampleEnum)

    assert ExampleEnum in f.type
    assert len(f.type) == 1


# This is meant to exercise `_seq_field`.
def test_pvector_field_enum_type():
    f = pvector_field(ExampleEnum)

    assert len(f.type) == 1
    assert ExampleEnum is list(f.type)[0].__type__
